"""
Allows the user to manipulate colors.

- rgb-hsl conversion (both ways).
- modifying colors by fading, lightening and darkening.
- obtaining Colour information (hue, saturation, light, etc).
- determining luminosity.

RGB is in readable format (#......) unless specified otherwise.
"""
import re

"""

import PerlLib.Colour

red, green, blue = Colour.rgb("mistyrose3")
hue, saturation, light = Colour.hsl("mistyrose3")

Colour.darken("mistyrose3", 20) # darken the Colour by 20%
Colour.lighten("mistyrose3", 20) # lighten the Colour by 20%

print(Colour.string("mistyrose3"))          # "#FF0000"
print(Colour.hsl_from_percent(0, 1, 0.5))   # "#FF0000"
"""

colour_list: dict[str, tuple[int, int, int]] = dict()

MAX_COLOUR = 255
LEN = 2


# --------------------------------------------------------
# __initialize_colours
# --------------------------------------------------------
def initialize_colours():
    """this function initializes all the colors"""
    global colour_list

    # if we have already done this, don't do it again
    if colour_list:
        return

    _colour_list = __colour_data.split("\n")
    for c in _colour_list:
        if not c:
            continue
        r, g, b, name = re.split(r"\s+", c.strip(), 3)
        if not name:
            continue
        name: str = name.strip()
        colour_list[name] = (int(r), int(g), int(b))


# --------------------------------------------------------
# _rgb2hsl
# Source:
# http://www.mathworks.com/matlabcentral/fileexchange/loadFile.do?objectId=3360&objectType=file#review_submission
# --------------------------------------------------------
def _rgb2hsl(r: float, g: float, b: float) -> tuple[float, float, float]:
    """
    Converts a percentage-based (0-1) RGB value to a percentage-based HSL value
    :param r: Red content of Colour.
    :param g: Green content of Colour.
    :param b: Blue content of Colour.
    :return: (h,s,l)
    """
    mini = min(r, g, b)
    maxi = max(r, g, b)

    # luminescence
    l = (maxi + mini) / 2

    # saturation
    if maxi == mini:
        return 0, 0, l
    elif l < 0.5:
        s = (maxi - mini) / (maxi + mini)
    else:
        s = (maxi - mini) / (2 - (maxi + mini))

    # hue (based on offset of maximum Colour)
    if r == maxi:
        h = (g - b) / (maxi - mini)
    elif g == maxi:
        h = 2 + (b - r) / (maxi - mini)
    else:
        h = 4 + (r - g) / (maxi - mini)

    h /= 6
    h = h + 1 if h < 0 else h
    return h, s, l


# --------------------------------------------------------
# _hsl2rgb
# Source:
# http://www.easyrgb.com/math.php?MATH=M19#text19
# --------------------------------------------------------
def _hsl2rgb(h: float, s: float, l: float) -> tuple[float, float, float]:
    """
    Converts a percentage-based (0-1) HSL value to a percentage-based RGB value
    :param h: Hue of Colour.
    :param s: Saturation of Colour.
    :param l: Luminescence of Colour.
    :return: (r,g,b)
    """
    if s == 0:
        return l, l, l
    elif l < 0.5:
        t1 = l * (s + 1)
    else:
        t1 = (l + s) - (s * l)

    t2 = 2 * l - t1

    r = _hue2rgb(t2, t1, h + (1 / 3))
    g = _hue2rgb(t2, t1, h)
    b = _hue2rgb(t2, t1, h - (1 / 3))
    return r, g, b


# --------------------------------------------------------
# _hue2rgb
# Source:
# http://www.easyrgb.com/math.php?MATH=M19#text19
# --------------------------------------------------------
def _hue2rgb(v1: float, v2: float, vH: float):
    vH += 1 if vH < 0 else 0
    vH -= 1 if vH > 1 else 0
    if (6 * vH) < 1:
        return v1 + (v2 - v1) * 6 * vH
    elif (2 * vH) < 1:
        return v2
    elif (3 * vH) < 2:
        return v1 + (v2 - v1) * ((2 / 3) - vH) * 6
    else:
        return v1


# --------------------------------------------------------
# string
# --------------------------------------------------------
def string(colour: str) -> str:
    """
    Returns the RGB value of a given Colour.
    - Parameter Colour -> The Colour to return.
    Supports giving a Colour name, as long as the Colour appears in the *nix 'showrgb' command.
    """
    global MAX_COLOUR
    global LEN
    if colour[0] != "#":
        colour_string: str = "#000000"
        if not colour_list:
            initialize_colours()

        if colour_list and colour in colour_list:
            colour_string = get_colour_string_from_rgb(
                colour_list[colour][0] / 255,
                colour_list[colour][1] / 255,
                colour_list[colour][2] / 255,
            )
    else:
        colour_string = colour

    LEN = int(len(colour_string[1::]) / 3)
    MAX_COLOUR = int("F" * LEN, 16)

    return colour_string


# --------------------------------------------------------
# setcolour_rgb
# --------------------------------------------------------
def get_colour_string_from_rgb(r: float, g: float, b: float) -> str:
    """
    Converts a percentage-based (0-1) RGB into a hex Colour string.
    :param r: Red content of Colour.
    :param g: Green content of Colour.
    :param b: Blue content of Colour.
    :return: hex_string
    """
    r = int(r * MAX_COLOUR + .5)
    b = int(b * MAX_COLOUR + .5)
    g = int(g * MAX_COLOUR + .5)

    return "#{:02X}{:02X}{:02X}".format(r, g, b)


# --------------------------------------------------------
# setcolour_hsl
# --------------------------------------------------------
def get_colour_string_from_hsl(h: float, s: float, l: float) -> str:
    """
    Converts a percentage-based (0-1) HSL into a readable RGB value.
    - Parameter h -> Hue of Colour.
    - Parameter s -> Saturation of Colour.
    - Parameter l -> Luminescence of Colour.
    """
    r, g, b = _hsl2rgb(h, s, l)
    c = get_colour_string_from_rgb(r, g, b)
    return c


# --------------------------------------------------------
# rgb
# --------------------------------------------------------
def rgb(colour: str) -> tuple[float, float, float]:
    """
    Converts a given Colour into a percentage-based RGB value.
    - Parameter Colour -> Name or RGB of given Colour.
    """
    r: float = int(string(colour)[1:LEN + 1:], 16) / MAX_COLOUR
    g: float = int(string(colour)[LEN + 1:LEN * 2 + 1:], 16) / MAX_COLOUR
    b: float = int(string(colour)[LEN * 2 + 1:LEN * 3 + 1:], 16) / MAX_COLOUR
    return r, g, b


# --------------------------------------------------------
# hsl
# --------------------------------------------------------
def hsl(colour: str) -> tuple[float, float, float]:
    """
    Converts a given Colour into a hue/saturation/light value
    - Parameter Colour -> Name or RGB of given Colour.
    """
    r, g, b = rgb(colour)
    return _rgb2hsl(r, g, b)


# --------------------------------------------------------
# darken
# --------------------------------------------------------
def darken(colour: str, amount: int = 10) -> str:
    """
    Darkens a Colour. Equivalent to lighten(Colour, -amount)
    - Parameter Colour -> The Colour to darken.
    - Parameter amount -> How much to darken by. Defaults to 10.
    """
    # get HSL
    h, s, l = hsl(string(colour))

    # reduce light
    l = l - amount / 100
    l = max(0.0, min(1.0, l))

    # convert to Colour string
    return get_colour_string_from_hsl(h, s, l)


# --------------------------------------------------------
# lighten
# --------------------------------------------------------
def lighten(colour: str, amount: int = 10) -> str:
    """
    Lightens a Colour. Equivalent to darken(Colour, -amount)
    - Parameter Colour -> The Colour to lighten.
    - Parameter amount -> How much to lighten by. Defaults to 10.
    """
    return darken(colour, -amount)


# --------------------------------------------------------
# is_light
# --------------------------------------------------------
def is_light(colour: str) -> bool:
    """
    Determines if the Colour is light or dark, based on visual luminosity
    - Parameter Colour -> The Colour to be determined
    """
    l = luminosity(string(colour))
    return l > 0.5


# --------------------------------------------------------
# luminosity
# --------------------------------------------------------
def luminosity(colour: str) -> float:
    """
    Returns the visual luminosity of a given Colour
    - Parameter Colour -> The Colour whose luminosity is returned
    """
    r, g, b = rgb(colour)
    return (r * 299 + g * 587 + b * 114) / 1437  # visual luminosity


# --------------------------------------------------------
# add
# --------------------------------------------------------
def add(c1: str, c2: str) -> str:
    """
    Adds one Colour to another, assuming that the light is shining through the colors

    White (clear) covered by black will return black

    Black (opaque) covered by white will return black

    50% grey covered by 50% grey will return 75% grey

    - Parameter c1 -> The first Colour to be added
    - Parameter c1 -> The second Colour to be added
    """
    c1 = string(c1)
    c2 = string(c2)

    r1, g1, b1 = rgb(c1)
    r2, g2, b2 = rgb(c2)

    # add & convert to Colour string
    return get_colour_string_from_rgb(r1 * r2, g1 * g2, b1 * b2)


__colour_data = """
199  21 133		medium violet red
176 196 222		light steel blue
102 139 139		paleturquoise4
159 121 238		mediumpurple2
141 182 205		lightskyblue3
  0 238 118		springgreen2
255 160 122		light salmon
154 205  50		yellowgreen
178  58 238		darkorchid2
 69 139 116		aquamarine4
 71  60 139		slateblue4
131 111 255		slateblue1
192 255  62		olivedrab1
139 105  20		goldenrod4
205 155  29		goldenrod3
142 229 238		cadetblue2
255 211 155		burlywood1
112 128 144		slategrey
255 228 225		mistyrose
 50 205  50		limegreen
224 255 255		lightcyan
218 165  32		goldenrod
220 220 220		gainsboro
135 206 255		skyblue1
240 255 240		honeydew
238 238   0		yellow2
205  79  57		tomato3
135 206 235		skyblue
 85  26 139		purple4
205 133   0		orange3
205 183 158		bisque3
238 213 183		bisque2
 87  87  87		grey34
252 252 252		gray99
161 161 161		gray63
112 112 112		gray44
 94  94  94		gray37
 84  84  84		gray33
 66  66  66		gray26
240 255 255		azure1
139 137 137		snow4
205 133  63		peru
219 112 147		pale violet red
139 129  76		lightgoldenrod4
 60 179 113		mediumseagreen
255 240 245		lavender blush
209  95 238		mediumorchid2
176 226 255		lightskyblue1
 72  61 139		darkslateblue
 25  25 112		midnightblue
255 160 122		lightsalmon1
255 250 205		lemonchiffon
173 255  47		green yellow
255 160 122		lightsalmon
240 128 128		light coral
 24 116 205		dodgerblue3
139  69   0		darkorange4
106  90 205		slate blue
 39  64 139		royalblue4
255  69   0		orange red
 50 205  50		lime green
224 255 255		light cyan
148   0 211		darkviolet
233 150 122		darksalmon
255 140   0		darkorange
 95 158 160		cadet blue
255  20 147		deep pink
238   0 238		magenta2
139  71  38		sienna4
238 230 133		khaki2
191 191 191		grey75
189 189 189		grey74
186 186 186		grey73
176 176 176		grey69
173 173 173		grey68
 89  89  89		grey35
 33  33  33		grey13
229 229 229		gray90
207 207 207		gray81
140 140 140		gray55
130 130 130		gray51
 79  79  79		gray31
238 233 233		snow2
205 145 158		pink3
 18  18  18		grey7
  3   3   3		gray1
139   0   0		red4
205   0   0		red3
210 180 140		tan
255   0   0		red
199  21 133		mediumvioletred
119 136 153		lightslategrey
135 206 250		lightskyblue
139  87  66		lightsalmon4
 34 139  34		forestgreen
 16  78 139		dodgerblue4
153  50 204		dark orchid
188 143 143		rosy brown
205 175 149		peachpuff3
124 205 124		palegreen3
238  64   0		orangered2
122 139 139		lightcyan4
139  58  58		indianred4
205  85  85		indianred3
  0   0 128		navyblue
105 105 105		dim grey
255  20 147		deeppink
139  76  57		salmon4
205 112  84		salmon3
196 196 196		grey77
145 145 145		gray57
 28  28  28		gray11
205 150 205		plum3
 23  23  23		gray9
 20  20  20		gray8
  0   0 139		blue4
245 245 220		beige
250 250 210		light goldenrod yellow
139 131 134		lavenderblush4
  0 206 209		dark turquoise
  0 206 209		darkturquoise
 47  79  79		darkslategrey
205 129  98		lightsalmon3
139 105 105		rosybrown4
255 228 225		misty rose
 78 238 148		seagreen2
205  92  92		indianred
255  20 147		deeppink1
  0   0 139		dark blue
230 230 250		lavender
253 245 230		oldlace
199 199 199		grey78
138 138 138		grey54
115 115 115		grey45
 54  54  54		grey21
247 247 247		gray97
245 245 245		gray96
242 242 242		gray95
224 224 224		gray88
222 222 222		gray87
219 219 219		gray86
179 179 179		gray70
 97  97  97		gray38
 31  31  31		gray12
250 240 230		linen
 72 209 204		medium turquoise
 72  61 139		dark slate blue
139 137 112		lemonchiffon4
193 255 193		darkseagreen1
205 192 176		antiquewhite3
186  85 211		mediumorchid
  0 255 127		springgreen
  0 134 139		turquoise4
 79 148 205		steelblue3
238 213 210		mistyrose2
209 238 238		lightcyan2
205  92  92		indian red
238  44  44		firebrick2
 65 105 225		royalblue
 95 158 160		cadetblue
108 166 205		skyblue3
205 205   0		yellow3
255 140 105		salmon1
139  90   0		orange4
255 105 180		hotpink
229 229 229		grey90
143 143 143		gray56
 99  99  99		gray39
 46  46  46		gray18
 36  36  36		gray14
139 102 139		plum4
 15  15  15		grey6
 15  15  15		gray6
205 173   0		gold3
255 215   0		gold1
  0   0 238		blue2
238 154  73		tan2
  0 255 255		cyan
  0 250 154		mediumspringgreen
188 238 104		darkolivegreen2
238 232 170		pale goldenrod
176 196 222		lightsteelblue
244 164  96		sandy brown
255 239 213		papaya whip
102 205   0		chartreuse3
139  34  82		violetred4
 67 110 238		royalblue2
 72 118 255		royalblue1
255 239 213		papayawhip
205 183 181		mistyrose3
224 255 255		lightcyan1
127 255 212		aquamarine
 74 112 139		skyblue4
139  58  98		hotpink4
205  96 144		hotpink3
238 106 167		hotpink2
169 169 169		darkgrey
105 105 105		dimgray
255  99  71		tomato
168 168 168		grey66
166 166 166		grey65
163 163 163		grey64
 84  84  84		grey33
 69  69  69		grey27
194 194 194		gray76
176 176 176		gray69
173 173 173		gray68
  0   0   0		grey0
240 255 255		azure
  0 250 154		medium spring green
139 101   8		darkgoldenrod4
205 149  12		darkgoldenrod3
238 173  14		darkgoldenrod2
184 134  11		darkgoldenrod
139  69  19		saddle brown
238 149 114		lightsalmon2
  0 104 139		deepskyblue4
  0 154 205		deepskyblue3
  0 178 238		deepskyblue2
  0 191 255		deepskyblue
255 127   0		darkorange1
205  50 120		violetred3
238  58 140		violetred2
255  62 150		violetred1
105  89 205		slateblue3
122 103 238		slateblue2
107 142  35		olive drab
255 106 106		indianred1
255  48  48		firebrick1
 83 134 139		cadetblue4
208  32 144		violetred
188 143 143		rosybrown
  0   0 128		navy blue
178  34  34		firebrick
139   0   0		dark red
255 255 255		grey100
139 126 102		wheat4
201 201 201		grey79
194 194 194		grey76
156 156 156		grey61
237 237 237		gray93
214 214 214		gray84
166 166 166		gray65
 92  92  92		gray36
 82  82  82		gray32
 33  33  33		gray13
 26  26  26		gray10
193 205 205		azure3
255 250 250		snow1
255 165  79		tan1
119 136 153		light slate gray
202 255 112		darkolivegreen1
100 149 237		cornflower blue
255 235 205		blanched almond
205 193 197		lavenderblush3
238 224 229		lavenderblush2
255 240 245		lavenderblush1
 85 107  47		darkolivegreen
255 240 245		lavenderblush
118 238 198		aquamarine2
208  32 144		violet red
179 238  58		olivedrab2
139 125 123		mistyrose4
255 228 225		mistyrose1
180 205 205		lightcyan3
240 128 128		lightcoral
127 255   0		chartreuse
255 218 185		peachpuff
152 251 152		palegreen
245 255 250		mintcream
126 192 238		skyblue2
255 228 181		moccasin
255  99  71		tomato1
205 105 201		orchid3
205  41 144		maroon3
250 128 114		salmon
207 207 207		grey81
158 158 158		grey62
 99  99  99		grey39
 97  97  97		grey38
 94  94  94		grey37
235 235 235		gray92
212 212 212		gray83
168 168 168		gray66
138 138 138		gray54
127 127 127		gray50
 77  77  77		gray30
 48  48  48		gray19
 38  38  38		gray15
131 139 139		azure4
  8   8   8		grey3
205 133  63		tan3
255 192 203		pink
190 190 190		gray
  0   0 255		blue
188 210 238		lightsteelblue2
202 225 255		lightsteelblue1
 32 178 170		light sea green
119 136 153		lightslategray
238 233 191		lemonchiffon2
  0 255 127		springgreen1
173 255  47		greenyellow
118 238   0		chartreuse2
112 128 144		slate grey
 58  95 205		royalblue3
176 224 230		powderblue
238 203 173		peachpuff2
144 238 144		palegreen2
245 255 250		mint cream
106  90 205		slateblue
238 229 222		seashell2
238  18 137		deeppink2
189 183 107		darkkhaki
139  28  98		maroon4
160  82  45		sienna
181 181 181		grey71
171 171 171		grey67
 46  46  46		grey18
150 150 150		gray59
110 110 110		gray43
 64  64  64		gray25
255 228 196		bisque
255   0   0		red1
123 104 238		mediumslateblue
255 236 139		lightgoldenrod1
238 221 130		light goldenrod
150 205 205		paleturquoise3
 96 123 139		lightskyblue4
  0 255 127		spring green
255 255 224		light yellow
245 245 245		white smoke
  0   0 205		medium blue
248 248 255		ghost white
 54 100 139		steelblue4
205 155 155		rosybrown3
255 218 185		peachpuff1
154 255 154		palegreen1
138  43 226		blueviolet
139 134 130		seashell4
169 169 169		darkgray
205 104  57		sienna3
102 102 102		grey40
232 232 232		gray91
209 209 209		gray82
 13  13  13		gray5
  0 238 238		cyan2
  0 255 255		cyan1
  0   0 255		blue1
255 250 250		snow
238 220 130		lightgoldenrod2
132 112 255		lightslateblue
180  82 205		mediumorchid3
105 139 105		darkseagreen4
  0 205 102		springgreen3
 34 139  34		forest green
108 123 139		slategray4
159 182 205		slategray3
185 211 238		slategray2
 65 105 225		royal blue
139 119 101		peachpuff4
 84 139  84		palegreen4
152 251 152		pale green
205  55   0		orangered3
255 193  37		goldenrod1
248 248 255		ghostwhite
139  26  26		firebrick4
205  38  38		firebrick3
122 197 205		cadetblue3
112 128 144		slategray
205 197 191		seashell3
193 205 193		honeydew3
139 136 120		cornsilk4
238 232 205		cornsilk2
155  48 255		purple1
105 105 105		dimgrey
139   0   0		darkred
255 246 143		khaki1
205 205 193		ivory3
179 179 179		grey70
153 153 153		grey60
 82  82  82		grey32
 56  56  56		grey22
 31  31  31		grey12
250 250 250		gray98
227 227 227		gray89
181 181 181		gray71
163 163 163		gray64
153 153 153		gray60
125 125 125		gray49
224 238 238		azure2
  8   8   8		gray3
187 255 255		paleturquoise1
171 130 255		mediumpurple1
147 112 219		medium purple
255 250 205		lemonchiffon1
  0 191 255		deep sky blue
205 179 139		navajowhite3
191  62 255		darkorchid1
255 140   0		dark orange
238 180  34		goldenrod2
189 183 107		dark khaki
238 118  33		chocolate2
238 197 145		burlywood2
240 255 240		honeydew1
  0 100   0		darkgreen
205 181 205		thistle3
238 210 238		thistle2
255 225 255		thistle1
  0   0 139		darkblue
216 191 216		thistle
238  48 167		maroon2
255  52 179		maroon1
135 135 135		grey53
112 112 112		grey44
 64  64  64		grey25
189 189 189		gray74
115 115 115		gray45
105 105 105		gray41
 89  89  89		gray35
 69  69  69		gray27
 59  59  59		gray23
 41  41  41		gray16
139  35  35		brown4
245 222 179		wheat
255 127  80		coral
139  90  43		tan4
250 250 210		lightgoldenrodyellow
132 112 255		light slate blue
 85 107  47		dark olive green
 47  79  79		dark slate gray
205 104 137		palevioletred3
 93  71 139		mediumpurple4
137 104 205		mediumpurple3
139  69  19		saddlebrown
176 224 230		powder blue
104  34 139		darkorchid4
154  50 205		darkorchid3
255 218 185		peach puff
105 139  34		olivedrab4
104 131 139		lightblue4
255 182 193		lightpink
211 211 211		lightgray
224 238 224		honeydew2
255 248 220		cornsilk1
253 245 230		old lace
255 130  71		sienna1
139 125 107		bisque4
218 112 214		orchid
205 198 115		khaki3
214 214 214		grey84
212 212 212		grey83
209 209 209		grey82
184 184 184		grey72
133 133 133		grey52
110 110 110		grey43
 66  66  66		grey26
 36  36  36		grey14
 26  26  26		grey10
191 191 191		gray75
135 135 135		gray53
 54  54  54		gray21
 51  51  51		gray20
205  51  51		brown3
 20  20  20		grey8
238   0   0		red2
  0   0 128		navy
190 190 190		grey
255 215   0		gold
102 205 170		mediumaquamarine
238 221 130		lightgoldenrod
 82 139 139		darkslategray4
155 205 155		darkseagreen3
180 238 180		darkseagreen2
139 131 120		antiquewhite4
250 235 215		antique white
  0 139  69		springgreen4
139 139 122		lightyellow4
255 250 240		floral white
127 255 212		aquamarine1
  0 197 205		turquoise3
 92 172 238		steelblue2
238 180 180		rosybrown2
255 182 193		light pink
211 211 211		light gray
238  99  99		indianred2
 30 144 255		dodgerblue
  0 100   0		dark green
 84 255 159		seagreen1
139  10  80		deeppink4
240 248 255		aliceblue
255   0 255		magenta1
255 105 180		hot pink
238 121  66		sienna2
255 131 250		orchid1
255 255 255		gray100
247 247 247		grey97
240 240 240		grey94
222 222 222		grey87
219 219 219		grey86
130 130 130		grey51
107 107 107		grey42
 48  48  48		grey19
240 240 240		gray94
217 217 217		gray85
156 156 156		gray61
238  59  59		brown2
240 230 140		khaki
  3   3   3		grey1
139 117   0		gold4
123 104 238		medium slate blue
 60 179 113		medium sea green
 47  79  79		dark slate grey
175 238 238		pale turquoise
175 238 238		paleturquoise
122  55 139		mediumorchid4
238 223 204		antiquewhite2
238 238 209		lightyellow2
144 238 144		light green
148   0 211		dark violet
233 150 122		dark salmon
127 255   0		chartreuse1
  0 245 255		turquoise1
244 164  96		sandybrown
255  69   0		orangered1
255 174 185		lightpink1
178 223 238		lightblue2
191 239 255		lightblue1
211 211 211		light grey
 46 139  87		seagreen4
 67 205 128		seagreen3
173 216 230		lightblue
205  16 118		deeppink3
169 169 169		dark grey
  0 139 139		dark cyan
222 184 135		burlywood
255 245 238		seashell
255 110 180		hotpink1
105 105 105		dim gray
  0 139 139		darkcyan
139 139   0		yellow4
255 255   0		yellow
160  32 240		purple
255 165   0		orange
139 139 131		ivory4
252 252 252		grey99
227 227 227		grey89
161 161 161		grey63
148 148 148		grey58
125 125 125		grey49
 79  79  79		grey31
 61  61  61		grey24
 51  51  51		grey20
  0 139   0		green4
  0 255   0		green1
186 186 186		gray73
171 171 171		gray67
205  91  69		coral3
238 106  80		coral2
238 174 238		plum2
139  99 108		pink4
255 255 240		ivory
 10  10  10		gray4
  5   5   5		gray2
238 201   0		gold2
102 205 170		medium aquamarine
119 136 153		light slate grey
205 190 112		lightgoldenrod3
162 205  90		darkolivegreen3
255 185  15		darkgoldenrod1
184 134  11		dark goldenrod
186  85 211		medium orchid
255 250 205		lemon chiffon
139 121  94		navajowhite4
  0 191 255		deepskyblue1
255 255 224		lightyellow
255 250 240		floralwhite
 30 144 255		dodger blue
  0   0 205		mediumblue
144 238 144		lightgreen
139  69  19		chocolate4
205 102  29		chocolate3
139 115  85		burlywood4
 64 224 208		turquoise
 70 130 180		steelblue
 46 139  87		sea green
124 252   0		lawngreen
131 139 131		honeydew4
169 169 169		dark gray
 46 139  87		seagreen
139  71 137		orchid4
255 231 186		wheat1
238 130 238		violet
255 255 240		ivory1
224 224 224		grey88
217 217 217		grey85
145 145 145		grey57
143 143 143		grey56
140 140 140		grey55
122 122 122		grey48
120 120 120		grey47
117 117 117		grey46
 77  77  77		grey30
 43  43  43		grey17
120 120 120		gray47
 74  74  74		gray29
238 169 184		pink2
 13  13  13		grey5
 10  10  10		grey4
  0 255   0		green
  0   0   0		gray0
165  42  42		brown
110 123 139		lightsteelblue4
110 139  61		darkolivegreen4
139  71  93		palevioletred4
135 206 250		light sky blue
121 205 205		darkslategray3
141 238 238		darkslategray2
151 255 255		darkslategray1
255 235 205		blanchedalmond
238 232 170		palegoldenrod
 25  25 112		midnight blue
 32 178 170		lightseagreen
205 201 165		lemonchiffon3
 47  79  79		darkslategray
154 205  50		yellow green
143 188 143		darkseagreen
250 235 215		antiquewhite
238 118   0		darkorange2
 69 139   0		chartreuse4
 70 130 180		steel blue
255 193 193		rosybrown1
154 205  50		olivedrab3
238 162 173		lightpink2
255  69   0		orangered
139 123 139		thistle4
135 206 235		sky blue
255 248 220		cornsilk
238 130  98		salmon2
238 122 233		orchid2
238 238 224		ivory2
237 237 237		grey93
235 235 235		grey92
232 232 232		grey91
 92  92  92		grey36
 74  74  74		grey29
 71  71  71		grey28
 41  41  41		grey16
201 201 201		gray79
199 199 199		gray78
196 196 196		gray77
122 122 122		gray48
 43  43  43		gray17
139  62  47		coral4
255 114  86		coral1
255 187 255		plum1
255 181 197		pink1
 23  23  23		grey9
  5   5   5		grey2
 18  18  18		gray7
  0 139 139		cyan4
  0   0 205		blue3
221 160 221		plum
100 149 237		cornflowerblue
164 211 238		lightskyblue2
255 239 219		antiquewhite1
238 207 161		navajowhite2
255 222 173		navajowhite1
205 205 180		lightyellow3
139   0 139		dark magenta
255 222 173		navajowhite
205 102   0		darkorange3
245 245 245		whitesmoke
  0 229 238		turquoise2
 99 184 255		steelblue1
139  95 101		lightpink4
154 192 205		lightblue3
124 252   0		lawn green
255 127  36		chocolate1
240 248 255		alice blue
107 142  35		olivedrab
211 211 211		lightgrey
210 105  30		chocolate
139   0 139		magenta4
205   0 205		magenta3
255 255   0		yellow1
125  38 205		purple3
145  44 238		purple2
238 154   0		orange2
255 165   0		orange1
255   0 255		magenta
255 228 196		bisque1
238 216 174		wheat2
176  48  96		maroon
139 134  78		khaki4
245 245 245		grey96
242 242 242		grey95
204 204 204		grey80
127 127 127		grey50
105 105 105		grey41
 38  38  38		grey15
 28  28  28		grey11
204 204 204		gray80
148 148 148		gray58
102 102 102		gray40
 87  87  87		gray34
 56  56  56		gray22
255  64  64		brown1
205 201 201		snow3
 72 209 204		mediumturquoise
162 181 205		lightsteelblue3
238 121 159		palevioletred2
255 130 171		palevioletred1
174 238 238		paleturquoise2
143 188 143		dark sea green
219 112 147		palevioletred
224 102 255		mediumorchid1
255 222 173		navajo white
147 112 219		mediumpurple
255 255 224		lightyellow1
 28 134 238		dodgerblue2
 30 144 255		dodgerblue1
139   0 139		darkmagenta
138  43 226		blue violet
102 205 170		aquamarine3
198 226 255		slategray1
112 128 144		slate gray
139  37   0		orangered4
205 140 149		lightpink3
173 216 230		light blue
153  50 204		darkorchid
152 245 255		cadetblue1
205 170 125		burlywood3
255 245 238		seashell1
205 200 177		cornsilk3
139  54  38		tomato4
238  92  66		tomato2
205 186 150		wheat3
250 250 250		grey98
150 150 150		grey59
 59  59  59		grey23
  0 205   0		green3
  0 238   0		green2
184 184 184		gray72
158 158 158		gray62
133 133 133		gray52
117 117 117		gray46
107 107 107		gray42
 71  71  71		gray28
 61  61  61		gray24
255 255 255		white
  0 205 205		cyan3
  0   0   0		black"""



from __future__ import annotations
from tkinter import *

from src.scheduling_and_allocation.modified_tk import AdvancedTreeview

mw = Tk()
mw.geometry("500x300")

frame = Frame(mw, bg='blue')
frame.pack(expand=1, fill="both")
tree = AdvancedTreeview(frame)
tree.pack(expand=1, fill='both')


def clear_tree():
    for kid in tree.get_children(""):
        tree.delete(kid)


def test_saving_objs():
    clear_tree()
    assert len(tree._item_to_obj_dict) == 0
    o1 = 'hello'
    o2 = 'goodbye'
    o3 = 'death to all'
    id1 = tree.insert("", any_object=o1, text=str(o1))
    id2 = tree.insert("", any_object=o2, text=str(o2))
    id3 = tree.insert("", any_object=o3, text=str(o3))
    id4 = tree.insert("", any_object=o1, text=str(o1))

    assert len(tree._item_to_obj_dict) == 4
    assert tree.get_obj_from_id(id1) == o1
    assert tree.get_obj_from_id(id2) == o2
    assert tree.get_obj_from_id(id3) == o3
    assert tree.get_obj_from_id(id4) == o1

    kids = tree.get_children("")
    assert kids[0] == id1
    assert kids[1] == id2
    assert kids[2] == id3
    assert kids[3] == id4

def test_deleting_items():
    clear_tree()
    assert len(tree._item_to_obj_dict) == 0
    o1 = 'hello'
    o2 = 'goodbye'
    o3 = 'death to all'
    id1 = tree.insert("", any_object=o1, text=str(o1))
    id2 = tree.insert("", any_object=o2, text=str(o2))
    id3 = tree.insert("", any_object=o3, text=str(o3))
    id4 = tree.insert("", any_object=o1, text=str(o1))

    assert len(tree._item_to_obj_dict) == 4

    tree.delete(id2, id3)
    assert len(tree._item_to_obj_dict) == 2
    assert tree.get_obj_from_id(id1) == o1
    assert tree.get_obj_from_id(id4) == o1


def test_insert_sorted():
    clear_tree()
    assert len(tree._item_to_obj_dict) == 0
    o1 = 'hello'
    o2 = 'goodbye'
    o3 = 'death to all'

    id2 = tree.insert_sorted("", any_object=o2, text=str(o2))
    id1 = tree.insert_sorted("", any_object=o1, text=str(o1))
    id3 = tree.insert_sorted("", any_object=o3, text=str(o3))
    assert len(tree._item_to_obj_dict) == 3

    kids = tree.get_children("")
    assert kids[0] == id3
    assert kids[1] == id2
    assert kids[2] == id1


def test_insert_sorted_but_not_sortable():
    clear_tree()
    assert len(tree._item_to_obj_dict) == 0
    o1 = UnSortableObject('hello')
    o2 = UnSortableObject('goodbye')
    o3 = UnSortableObject('death to all')

    id2 = tree.insert_sorted("", any_object=o2, text=str(o2))
    id1 = tree.insert_sorted("", any_object=o1, text=str(o1))
    id3 = tree.insert_sorted("", any_object=o3, text=str(o3))

    assert len(tree._item_to_obj_dict) == 3

    kids = tree.get_children("")
    assert kids[0] == id2
    assert kids[1] == id1
    assert kids[2] == id3


class UnSortableObject:
    def __init__(self, a: str):
        self.a = a
    def __str__(self):
        return self.a

class SortableObject:
    def __init__(self, a: str):
        self.a = a

    def __lt__(self, other: SortableObject):
        return self.a < other.a

    def __str__(self):
        return self.a




from __future__ import annotations
import sys
from os import path
import pytest

from src.scheduling_and_allocation.model import SemesterType, Course, InvalidSectionNumberForCourseError


class Lab:
    def __lt__(self, other: Lab):
        return True


class Stream:
    def __lt__(self, other: Stream):
        return True


class Teacher:
    def __init__(self):
        self.firstname = ""
        self.lastname = ""
    def __lt__(self, other: Teacher):
        return True


def test_full_constructor():
    """Verifies that constructor parameters are applied correctly"""
    num = "102-NYA-043"
    name = "My Course"
    sem = SemesterType.summer
    alloc = False
    c = Course(num, name, sem, 3.0, alloc)
    assert c.name == name
    assert c.number == num
    assert c.semester == SemesterType.validate(sem)
    assert c.needs_allocation == alloc
    assert c.hours_per_week == 3.0


def test_name_getter():
    """Verifies that name getter works as intended."""
    course = Course("420-101", "Intro to Programming")
    assert course.name == "Intro to Programming"


def test_name_setter():
    """Verifies that name setter works as intended."""
    course = Course("420-101")
    name = "Intro to Programming"
    course.name = name
    assert name == course.name


def test_number_getter():
    """Verifies that the number getter works as intended."""
    course = Course("420-101")
    x=course.number
    assert course.number == '420-101'


def test_add_section_returns_section():
    """Verifies that add_section returns a section"""
    course_2 = Course("420-101")
    section = course_2.add_section(section_id=1)
    assert len(course_2.sections()) == 1
    assert section.course == course_2


def test_add_section_good():
    """Verifies that add_section can add a valid Section to this Course, and that the Course is
    added to the Section itself. """
    course_2 = Course("420-101")
    section = course_2.add_section(section_id=1)
    assert len(course_2.sections()) == 1
    assert section.course == course_2


def test_add_section_duplicate():
    """Verifies that add_section() raises an Exception when trying to add a duplicate of an
    existing Section to the Course. """
    course_1 = Course("abc")
    course_1.add_section(number="dupl", section_id=1)
    with pytest.raises(InvalidSectionNumberForCourseError) as e:
        course_1.add_section(number="dupl", section_id=2)
    assert "section number is not unique" in str(e.value).lower()


def test_get_section_good():
    """Verifies that get_section_by_number() returns an existing section from this Course."""
    course = Course("abc")
    num = "420.AO"
    section = course.add_section(num, section_id=1)
    assert course.get_section_by_number(num).id == section.id


def test_get_section_bad():
    """Verifies that get_section_by_number() doesn't crash the program when trying to get_by_id a Section
    that doesn't exist. """
    course = Course("abc")
    bad_num = "420"
    assert course.get_section_by_number(bad_num) is None


def test_get_section_by_id_good():
    """Verifies that get_section_by_id() works when receiving a valid Section ID as input."""
    course = Course("abc")
    section = course.add_section(section_id=1)
    assert course.get_section_by_id(1).id == section.id


def test_get_section_by_id_bad():
    """Verifies that get_section_by_id() won't crash the program when given a bad ID as input,
    returning None instead. """
    course = Course("abc")
    course.add_section(section_id=1)
    bad_id = 999
    assert course.get_section_by_id(bad_id) is None


def test_get_section_by_name_good():
    """Verifies that get_sections_by_name() returns a list with the correct section when given
    a valid name as input. """
    course = Course("abc")
    name = "test"
    course.add_section("a", name)
    course.add_section("b", name)
    course.add_section("c", "x")
    course.add_section("d", "y")
    sections = course.get_sections_by_name(name)
    assert len(sections) == 2 and "a" in (s.number for s in sections) and "b" in (s.number for s in sections)


def test_get_section_by_name_bad():
    """Verifies that get_sections_by_name() returns an empty tuple when given an invalid
    section name. """
    course = Course("abc")
    course.add_section("",  "test")
    bad_name = "foo"
    sections = course.get_sections_by_name(bad_name)
    assert len(sections) == 0


def test_remove_section_good():
    """Verifies that remove_section() works as intended when asked to remove a legitimate
    Section. """
    course = Course("abc")
    section1 = course.add_section("a")
    section2 = course.add_section("b")
    course.remove_section(section2)
    assert len(course.sections()) == 1
    assert course.sections()[0].id == section1.id


def test_sections():
    """Verifies that sections() returns a list of all the Sections assigned to this Course."""
    course = Course("abc")
    section1 = course.add_section("420", section_id=1)
    section2 = course.add_section("555", section_id=2)
    sections = course.sections()
    assert len(sections) == 2
    assert section1 is not None and section1 in sections
    assert section2 is not None and section2 in sections


def test_number_of_sections():
    """Verifies that number_of_sections() correctly returns the number of Sections assigned
    to this Course. """
    course = Course("abc")
    course.add_section("420", section_id=1)
    course.add_section("555", section_id=2)
    assert course.number_of_sections() == len(course.sections())


def test_sections_for_teacher():
    """Verifies that get_sections_for_teacher() returns a list of all sections featuring this
    teacher in this course. """
    # Note, sections have to have blocks, or this won't work
    course = Course("abc")
    section_1 = course.add_section("420")
    section_2 = course.add_section("421")
    section_1.add_block()
    section_2.add_block()
    teach = Teacher()
    section_1.add_teacher(teach)
    teach_sections = course.get_sections_for_teacher(teach)
    assert section_1 in teach_sections
    assert section_2 not in teach_sections


def test_sections_for_teacher_empty():
    """Verifies that get_sections_for_teacher() returns an empty list if no Teacher has been
    assigned to the Course. """
    course = Course("abc")
    course.add_section("420")
    course.add_section("421")
    teach2 = Teacher()
    teach_sections = course.get_sections_for_teacher(teach2)
    assert len(teach_sections) == 0


def test_has_teacher():
    """Verifies that has_teacher() returns True if the specified Teacher has been assigned to
    this course. """
    # NOTE: section has to have blocks for this to work
    course = Course("abc", "Course 1")
    section = course.add_section("420", "Section 1")
    section.add_block()
    teacher = Teacher()
    section.add_teacher(teacher)
    assert course.has_teacher(teacher)


def test_does_not_have_teacher():
    """Verifies that has_teacher() returns False if the specified Teacher has not been assigned to
    this course. """
    course = Course("abc", "Course 1")
    section = course.add_section("420", "Section 1")
    teacher = Teacher()
    section.add_teacher(teacher)
    teacher_not_assigned = Teacher()
    assert not course.has_teacher(teacher_not_assigned)


def test_blocks_good():
    """Verifies that blocks() returns a list of all Blocks that have been assigned to this
    Course. """
    course = Course("abc")
    section = course.add_section("420")
    block1 = section.add_block()
    block2 = section.add_block()
    blocks = course.blocks()
    assert len(blocks) == 2 and block1 in blocks and block2 in blocks


def test_blocks_bad():
    """Verifies that blocks() returns an empty list when the Sections of this Course contain
    no Blocks. """
    course = Course("abc")
    course.add_section("420", section_id=1)
    blocks = course.blocks
    assert len(blocks()) == 0


def test_str_representation_full():
    """Too difficult to set_default_fonts_and_colours up, and not an issue (for now) if the wrong string is returned"""


def test_description():
    """Verifies that title prints a brief string containing the Course's
    number and name. """
    course = Course("abc", "Course 1")
    description = course.title
    assert f"{course.number}: {course.name}" in description


def test_streams():
    """Verifies that stream_ids() returns a list of all Streams belonging to the Sections of
    this Course. """
    course = Course("abc", "Course 1")
    section = course.add_section("420",  "Section 1")
    stream = Stream()
    section.add_stream(stream)
    streams = course.streams
    assert len(streams()) == 1 and stream in streams()


def test_streams_empty():
    """Verifies that stream_() returns an empty list if no Streams are assigned to this
    Course. """
    course = Course("abc", "Course 1")
    course.add_section("420",  "Section 1")
    streams = course.streams
    assert len(streams()) == 0


def test_has_stream():
    """Verifies that has_stream_with_id() returns true if the passed Stream exists within the Course."""
    course = Course("abc", "Course 1")
    section = course.add_section("420",  "Section 1")
    stream = Stream()
    section.add_stream(stream)
    assert course.has_stream(stream) is True


def test_has_stream_false():
    """Verifies that has_stream() returns false if the Course doesn't contain the passed
    Stream. """
    course = Course("abc", "Course 1")
    course.add_section("420",  "Section 1")
    bad_stream = Stream()
    assert course.has_stream(bad_stream) is False


def test_has_stream_false_doesnt_fail_with_no_sections():
    """Verifies that has_stream() returns false if the Course doesn't contain the passed
    Stream. """
    course = Course("abc", "Course 1")
    bad_stream = Stream()
    assert course.has_stream(bad_stream) is False


def test_assign_teacher_good():
    """Verifies that assign_teacher assigns the passed Teacher to all Sections of the
    Course. """
    # Note: sections requires blocks for this to work
    course = Course("abc", "Course 1")
    section1 = course.add_section("420",  "Section 1")
    section2 = course.add_section("421",  "Section 1")
    section1.add_block()
    section2.add_block()
    teacher = Teacher()
    course.add_teacher(teacher)
    teachers = section1.teachers
    assert len(teachers()) == 1 and teacher in teachers()
    teachers = section2.teachers
    assert len(teachers()) == 1 and teacher in teachers()


def test_assign_lab_good():
    """Verifies that assign_lab can assign a legitimate Lab to all Sections of the Course."""
    course = Course("abc", "Course 1")
    # NOTE: labs are only saved if we have blocks
    section = course.add_section("420",  "Section 1")
    section2 = course.add_section("421",  "Section 1")
    section.add_block()
    section2.add_block()
    lab = Lab()
    course.add_lab(lab)
    assert lab in section.labs()
    assert lab in section2.labs()


def test_assign_stream_good():
    """Verifies that assign_stream can successfully assign a valid Stream to all Sections
    of the Course. """
    course = Course("abc", "Course 1")
    section = course.add_section("420",  "Section 1")
    section2 = course.add_section("421", "Section 1")

    stream = Stream()
    course.add_stream(stream)
    assert len(section.streams()) == 1 and stream in section.streams()
    assert len(section2.streams()) == 1 and stream in section2.streams()


def test_remove_teacher_good():
    """Verifies that remove_teacher can successfully remove the specified Teacher from the
    Course. """
    course = Course("abc", "Course 1")
    teacher_1 = Teacher()
    teacher_2 = Teacher()
    # TODO: This is a bad design because I have to create a section before the teacher is assigned to it.
    section = course.add_section()
    section.add_block()
    course.add_teacher(teacher_1)
    course.add_teacher(teacher_2)
    course.remove_teacher(teacher_1)
    assert len(course.teachers()) == 1 and teacher_1 not in course.teachers()


def test_remove_all_teachers():
    """Verifies that remove_all_teachers() works as intended."""
    course = Course("abc", "Course 1")
    section1 = course.add_section("420",  "Section 1")
    section2 = course.add_section("421",  "Section 1")
    section1.add_block()
    section2.add_block()
    teacher_1 = Teacher()
    teacher_2 = Teacher()
    teacher_3 = Teacher()

    course.add_teacher(teacher_1)
    course.add_teacher(teacher_2)
    course.add_teacher(teacher_3)
    assert len(section1.teachers()) == 3
    assert len(course.teachers()) == 3
    course.remove_all_teachers()
    assert len(course.teachers()) == 0
    assert len(section1.teachers()) == 0
    assert len(section2.teachers()) == 0


def test_remove_stream_good():
    """Verifies that remove_stream() works as intended."""
    course = Course("abc", "Course 1")
    course.add_section("420",  "Section 1")
    course.add_section("421",  "Section 1")

    stream_1 = Stream()
    stream_2 = Stream()
    course.add_stream(stream_1)
    course.add_stream(stream_2)
    course.remove_stream(stream_1)
    assert len(course.streams()) == 1 and stream_1 not in course.streams() and stream_2 in course.streams()


def test_remove_all_streams():
    """Verifies that remove_all_streams() works as intended."""
    course = Course("abc", "Course 1")
    course.add_section("420",  "Section 1")
    stream_1 = Stream()
    stream_2 = Stream()
    course.add_stream(stream_1)
    course.add_stream(stream_2)
    course.remove_all_streams()
    assert len(course.streams()) == 0



PREP_FACTOR                = 0.9
PREP_BONUS_LIMIT           = 3.0
PREP_BONUS_FACTOR          = 0.2  # 1.1 - 0.9
PREP_CRAZY_BONUS_FACTOR    = 1.0

PES_FACTOR                 = 0.04
PES_BONUS_LIMIT            = 415
PES_BONUS_FACTOR           = 0.03

STUDENT_BONUS_LIMIT        = 75
STUDENT_BONUS_FACTOR       = 0.01
STUDENT_CRAZY_BONUS_LIMIT  = 160
STUDENT_CRAZY_BONUS_FACTOR = 0.1

HOURS_FACTOR               = 1.2

CI_FTE_PER_SEMESTER        = 40

prep_hours = 0
pes = 0
total_students = 0
num_preps = 0
hours = 0

while True:
    course_name = input("\nEnter Course name ").strip()
    if course_name == "":
        break
    new_prep_hours = int(input("How many hours? "))
    sections = int(input("Number of sections? "))
    students = int(input("Total number of students for all sections? "))

    prep_hours += new_prep_hours
    pes = pes + new_prep_hours * students
    hours += new_prep_hours * sections
    total_students += students
    num_preps += 1

# ------------------------------------------------------------------------
# hours in classroom * number of students
# ------------------------------------------------------------------------
CI_student = pes * PES_FACTOR

# bonus if pes is over 415
surplus = pes - PES_BONUS_LIMIT
bonus = surplus * PES_BONUS_FACTOR if surplus > 0 else 0
CI_student += bonus

# another bonus if total number of students is over student_bonus_limit
# (only for courses over 3 hours)
bonus = total_students * STUDENT_BONUS_FACTOR if total_students >= STUDENT_BONUS_LIMIT else 0
CI_student += bonus

# and yet another bonus if number of students is over Crazy student limit
if total_students >= STUDENT_CRAZY_BONUS_LIMIT:
    bonus = (total_students - STUDENT_CRAZY_BONUS_LIMIT) * STUDENT_CRAZY_BONUS_FACTOR
    CI_student += bonus

# ------------------------------------------------------------------------
# Preps (based on # of prep hours PER course)
# ------------------------------------------------------------------------
CI_preps = prep_hours * PREP_FACTOR

# bonus if number is the PREP_BONUS_LIMIT
if num_preps == PREP_BONUS_LIMIT:
    CI_preps += prep_hours * PREP_BONUS_FACTOR

# more bonus if over the limit
elif num_preps > PREP_BONUS_LIMIT:
    CI_preps += prep_hours * PREP_CRAZY_BONUS_FACTOR

# ------------------------------------------------------------------------
# Hours (based on contact hours per week)
# ------------------------------------------------------------------------
CI_hours = hours * HOURS_FACTOR

# ------------------------------------------------------------------------
# all done
# ------------------------------------------------------------------------
print( f"\nCI: \n"
       f"prep:               {CI_preps:5.2f}  \n"
       f"hours in classroom: {CI_hours:5.2f}  \n"
       f"student*hours:      {CI_student:5.2f}")
print( f"Total CI:           {CI_hours + CI_preps + CI_student:5.2f}")



import csv
import math
from typing import Optional

from ..model import Schedule, Section, Block, Teacher


def time_string(hour) -> str:
    minutes, hours = math.modf(hour)
    return f"{int(hours)}{int(round(minutes*60, 0)):02d}"


def registrars_output(schedule: Schedule, filename: str):
    headers = ["Course Name",
               "Course No.",
               "Section",
               "Start Time",
               "End Time",
               "Day",
               "Teacher Last Name",
               "Teacher First Name",
               "Room",
               "Other Rooms Used"
               ]
    fh = open(filename, "w")

    csv_file = csv.writer(fh)
    csv_file.writerow(headers)

    for course in (c for c in schedule.courses() if c.needs_allocation):
        for section in course.sections():

            # non classroom teaching
            if len(section.blocks()) == 0:
                row = [course.name, course.number, section.number, "","", ""]
                rows = add_teachers(section.section_defined_teachers(), row)
                for row in rows:
                    add_labs(section, row)
                    csv_file.writerow(row)
                continue

            # classroom teaching
            for block in section.blocks():
                row = [course.name, course.number, section.number, time_string(block.start),
                       time_string(block.start + block.duration), block.day.name.capitalize()]
                rows = add_teachers(block.teachers(), row)
                for row in rows:
                    add_labs(section, row)
                    csv_file.writerow(row)
    return ""

def add_labs(object: Section|Block, initial_row: list):
    labs = object.labs()
    if len(labs) == 0:
        initial_row.extend(["",""])
    elif len(labs) == 1:
        initial_row.extend([labs[0].number,""])
    else:
        initial_row.extend([labs[0].number, " ".join((l.number for l in labs[1:]))])

def add_teachers(teachers: tuple[Teacher, ...], initial_row: list) -> list[list[str]]:
    rows = []
    if len(teachers) == 0:
        rows.append(["",""])
    for teacher in teachers:
        rows.append([teacher.lastname, teacher.firstname])
    updated_rows = []
    for row in rows:
        updated_rows.append([*initial_row])
        updated_rows[-1].extend(row)
    return updated_rows




from functools import partial
from tkinter import *

from src.scheduling_and_allocation.gui_dialogs.add_section_dialog_tk import AddSectionDialogTk
from src.scheduling_and_allocation.gui_dialogs.edit_section_dialog_tk import EditSectionDialogTk
from src.scheduling_and_allocation.model import Schedule, Course, Section, WeekDay, SemesterType

mw=Tk()
frame = Frame(mw)
frame.pack()
mw.geometry("400x400")

schedule = Schedule()

# teachers
t1 = schedule.add_update_teacher("Jane", "Doe", "0.25", teacher_id="1")
t2 = schedule.add_update_teacher("John", "Doe", teacher_id="2")
t3 = schedule.add_update_teacher("Babe", "Ruth", teacher_id="3")
t4 = schedule.add_update_teacher("Bugs", "Bunny", teacher_id="4")

# labs
l1 = schedule.add_update_lab("P107", "C-Lab")
l2 = schedule.add_update_lab("P322", "Mac Lab")
l3 = schedule.add_update_lab("P325")
l4 = schedule.add_update_lab("BH311", "Britain Hall")

# streams
st1 = schedule.add_update_stream("1A", "Math Stream")
st2 = schedule.add_update_stream("1B")

# courses
c_001:Course = schedule.add_update_course("001", "Basket Weaving", SemesterType.fall)
s_001_1:Section = c_001.add_section("1", section_id=1)
s_001_1.add_stream(st1)
b_001_1_1 = s_001_1.add_block(WeekDay.Monday, 8)
b_001_1_2 = s_001_1.add_block(WeekDay.Monday, 10)
b_001_1_1.add_teacher(t1)
b_001_1_2.add_teacher(t1)
b_001_1_1.add_lab(l1)
b_001_1_2.add_lab(l1)

s_001_2 = c_001.add_section("2", section_id=2)
s_001_2.add_stream(st2)
b_001_2_1 = s_001_2.add_block(WeekDay.Tuesday, 8)
b_001_2_2 = s_001_2.add_block(WeekDay.Tuesday, 10)
b_001_2_1.add_teacher(t2)
b_001_2_2.add_teacher(t3)
b_001_2_1.add_lab(l2)
b_001_2_2.add_lab(l3)


def go_edit():
    title = f"{s_001_1.course.name} ({s_001_1.hours} hrs)"
    text = s_001_1.title
    block_data = []
    for b in s_001_1.blocks():
        block_data.append( (b.day.name, str(b.start), str(b.duration)))
    db = EditSectionDialogTk(frame,
                             course_description = title,
                             section_description = text,
                             assigned_teachers =[t1, t2, t3],
                             non_assigned_teachers = [t4],
                             assigned_labs = [l2],
                             non_assigned_labs = [l1, l3, l4],
                             assigned_streams = [],
                             non_assigned_streams = [st1,st2],
                             current_blocks= block_data,
                             apply_changes=partial(schedule.update_section, s_001_1),
                             course_hours = s_001_1.course.hours_per_week)

def go_add():
    pass
    title = f"{s_001_1.course.name} ({s_001_1.hours} hrs)"
    db = AddSectionDialogTk(frame,
                             course_description = title,
                             apply_changes=partial(schedule.add_sections,s_001_1.course),
                             course_hours = s_001_1.course.hours_per_week),

Button(frame, text="Edit Section", command=go_edit).pack()
Button(frame, text="Add Section", command=go_add).pack()
mw.mainloop()


from __future__ import annotations
from tkinter import *
from tkinter import ttk
from typing import *
global img_open, img_close, img_empty



class AdvancedTreeview(ttk.Treeview):
    """
    A subclass of ttk.TreeView object that will keep track of which object belongs with
    which Treeview item.

    see: https://docs.python.org/3/library/tkinter.ttk.html#treeview
    """

    def __init__(self, root: Frame, **kwargs):
        self._item_to_obj_dict = dict()
        super().__init__(root, **kwargs)

    def open_branch(self, tree_id: str):
        """Opens all the leafs in this branch"""
        self.item(tree_id, open=True)
        for leaf in self.get_children(tree_id):
            self.open_branch(leaf)

    #
    def insert(self, parent_id: str,
               any_object: Optional[Any],
               *,
               index: int | Literal['end'] = 'end', obj_id: str = None,
               **options) -> str:
        """
        override the ttk.treeview.insert command, adding 'any_object' as a new parameter

        Creates a new item and return the item identifier of the newly created item.

        see: https://docs.python.org/3/library/tkinter.ttk.html#treeview

        """

        obj_id = super().insert(parent_id, index, obj_id, **options)
        self._item_to_obj_dict[obj_id] = any_object
        return obj_id

    def insert_sorted(self, parent_id: str, any_object: Optional[Any], **options):
        """
        If the child objects support sorting, and if all siblings are of the
        same resource_type, then insert in sorted order, else put at end of siblings

        For allowed options, see: https://docs.python.org/3/library/tkinter.ttk.html#treeview
        """

        siblings = self.get_children(parent_id)

        sortable: bool = hasattr(any_object, "__lt__")

        # find index where to insert
        # NOTE: Assumes that 'get_children' returns items in row order
        options['index'] = 'end'
        if sortable:
            # use try except because maybe siblings cannot be compared to one another
            try:
                for index, kid_id in enumerate(siblings):
                    if any_object < self._item_to_obj_dict[kid_id]:
                        options['index'] = index
                        break
            except AttributeError:
                pass
            except TypeError:
                pass

        # insert
        return self.insert(parent_id, any_object=any_object, **options)

    def get_obj_from_id(self, obj_id: str) -> Optional[Any]:
        """Gets and returns the object associated with the tree item, unless 'obj_id' is
        not valid, in which case it returns None"""
        if obj_id in self._item_to_obj_dict.keys():
            return self._item_to_obj_dict[obj_id]
        else:
            return None

    def delete(self, *item_ids):
        """Delete all specified items and all their descendants. The root item may not be deleted."""

        # remove item, and all its descendants from the dictionary of objects
        for obj_id in item_ids:
            if self.exists(obj_id):
                self._prune(obj_id)
                super().delete(obj_id)

    def expand_whole_branch(self, tree_id: str):
        """Opens all the twigs in this branch"""
        self.item(tree_id, open=True)
        for twig in self.get_children(tree_id):
            self.expand_whole_branch(twig)


    def _prune(self, obj_id: str):
        """Not only remove the tree branch, but remove the data from tree_objects as well"""
        if self.exists(obj_id):

            for twig in self.get_children(obj_id):
                self._prune(twig)
            self._item_to_obj_dict.pop(obj_id, None)


"""
# =================================================================
# footer
# =================================================================

=head1 AUTHOR

Sandy Bultena

=head1 COPYRIGHT

Copyright (c) 2023 Sandy Bultena 

All Rights Reserved.

This module is free software. It may be used, redistributed
and/or modified under the terms of the Perl Artistic License

     (see http://www.perl.com/perl/misc/Artistic.html)

=cut
"""



"""Calculate the CI for a given teacher"""
from __future__ import annotations
from .ci_constants import *
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..model.schedule import Schedule
    from ..model.course import Course
    from ..model.teacher import Teacher

# for debugging
print_flag = False


def debug(*args):
    if not print_flag:
        return
    print(args)

def calculate_ci(teacher:Teacher, schedule: Schedule):
    c = CICalc(teacher, schedule)
    return c.calculate()

class CICalc:
    def __init__(self, teacher:Teacher, schedule: Schedule):
        """
        :param teacher:
        :parm schedule:
        """
        self._reset()
        self.prep_hours = 0
        self.schedule = schedule
        self.teacher = teacher
        self.release = 0

    def _reset(self):
        self.pes = 0
        self.num_preps = 0
        self.hours = 0
        self.students = 0
        self.ntu_students = 0
        self.release = 0
        self.dirty_flag = False

    def calculate(self, ) -> float:
        """
        Calculate the CI for the teacher and schedule specified during the init
        """
        schedule = self.schedule
        teacher = self.teacher

        courses: list[Course] = [c for c in schedule.courses() if c.has_allocated_teacher(teacher)]

        self._reset()
        self.release = teacher.release or 0

        debug("")
        debug("")
        debug(f"--------- {teacher.firstname}---------")
        debug("courses list")
        for course in courses:
            debug(f"{course.number} {course.name}")
        debug("end list")

        # per course
        for course in courses:
            max_prep_hours = 0
            debug(f"*********** {course.name}")
            hours = 0

            # per section
            for section in course.get_sections_for_allocated_teacher(teacher):
                debug(f"   Section: {section}")
                hours = section.get_teacher_allocation(teacher)
                students = section.num_students

                debug(f"{course.name}, Section: {section.number}, hours: {hours}")

                max_prep_hours = max_prep_hours if max_prep_hours > hours else hours

                self.pes = self.pes + hours * students
                self.students += students
                # Perl ver includes commented hours >= 3 check
                self.ntu_students += students
                self.hours += hours

            self.prep_hours += hours
            self.num_preps += 1

        # return
        total = self._total()
        debug(f"CI {teacher}: {total}")
        return total

    def _total(self) -> float:

        # ------------------------------------------------------------------------
        # PES (based on # of students and contact hours)
        # ------------------------------------------------------------------------
        CI_student = self.pes * PES_FACTOR
        debug(f"PES: {self.pes}\nStudent factor: {CI_student}")

        # bonus if pes is over 415
        surplus = self.pes - PES_BONUS_LIMIT
        bonus = surplus * PES_BONUS_FACTOR if surplus > 0 else 0
        debug(f"Bonus PES > 415: {bonus}")
        CI_student += bonus

        # another bonus if total number of students is over student_bonus_limit
        # (only for courses over 3 hours)
        debug(f"Students {self.students}")
        bonus = self.ntu_students * STUDENT_BONUS_FACTOR if self.ntu_students >= STUDENT_BONUS_LIMIT else 0
        debug(f"Bonus students over 75: {bonus}")
        CI_student += bonus

        # and yet another bonus if number of students is over Crazy student limit
        if self.ntu_students >= STUDENT_CRAZY_BONUS_LIMIT:
            bonus = (self.ntu_students - STUDENT_CRAZY_BONUS_LIMIT) * STUDENT_CRAZY_BONUS_FACTOR
            CI_student += bonus

        # ------------------------------------------------------------------------
        # Preps (based on # of prep hours PER course)
        # ------------------------------------------------------------------------
        CI_preps = self.prep_hours * PREP_FACTOR
        debug(f"Prep Hours {self.prep_hours}\nCI prep: {CI_preps}")

        # bonus if number is the PREP_BONUS_LIMIT
        if self.num_preps == PREP_BONUS_LIMIT:
            CI_preps += self.prep_hours * PREP_BONUS_FACTOR
            debug(f"CI bonus prep: {self.prep_hours * PREP_BONUS_FACTOR}")

        # more bonus if over the limit
        elif self.num_preps > PREP_BONUS_LIMIT:
            CI_preps += self.prep_hours * PREP_CRAZY_BONUS_FACTOR
            debug(f"CI bonus bonus prep: {self.prep_hours * PREP_CRAZY_BONUS_FACTOR}")

        # ------------------------------------------------------------------------
        # Hours (based on contact hours per week)
        # ------------------------------------------------------------------------
        CI_hours = self.hours * HOURS_FACTOR
        debug(f"CI hours: {CI_hours}")

        # ------------------------------------------------------------------------
        # Release
        # ------------------------------------------------------------------------
        CI_release = self.release * CI_FTE_PER_SEMESTER
        debug(f"CI release: {CI_release}")

        # ------------------------------------------------------------------------
        # all done
        # ------------------------------------------------------------------------
        return float(CI_release) + float(CI_hours) + float(CI_preps) + float(CI_student)



"""Dialog box to edit a 'block' object"""
from __future__ import annotations

import tkinter as tk

from tkinter.simpledialog import Dialog
from typing import Callable, Literal
from ..gui_generics.number_validations import  entry_float, entry_int, validate_float, validate_int

from ..gui_generics.add_remove_tk import AddRemoveTk

# =====================================================================================================================
# Edit an Existing Block
# =====================================================================================================================
class AddEditBlockDialogTk(Dialog):
    def __init__(self, frame:tk.Frame,
                 add_edit_type: Literal['add', 'edit'],
                 duration: float,
                 assigned_teachers: list,
                 non_assigned_teachers: list,
                 assigned_labs: list,
                 non_assigned_labs: list,
                 apply_changes: Callable[[int,float,list,list], None]):
        """
        :param frame: the frame which will be the parent of the dialog box
        :param add_edit_type: are we adding a new block or editing an existing bloc
        :param duration: how long is the block (class duration)
        :param assigned_teachers: who has been assigned to this block
        :param non_assigned_teachers: who has not been assigned to this block
        :param assigned_labs: labs that have been reserved for this block
        :param non_assigned_labs: labs that have not been assigned to this block
        :param apply_changes: who ya gonna call (ghost busters) once the user clicks ok
        """

        self.add_edit_type = add_edit_type
        self.old_duration = duration
        self.top_frame = frame
        self._assigned_teachers = assigned_teachers
        self._assigned_labs = assigned_labs
        self._non_assigned_teachers = non_assigned_teachers
        self._non_assigned_labs = non_assigned_labs
        self._apply_changes = apply_changes

        self.tk_duration = tk.StringVar(value=str(duration))
        self.tk_new_blocks = tk.StringVar(value="1")
        if add_edit_type == 'edit':
            self.tk_new_blocks.set("0")

        title = "Edit Class Time(s)" if add_edit_type == 'edit' else "Add Class"
        super().__init__(frame.winfo_toplevel(), title)

    # ================================================================================================================
    # The content of the main body of the dialog box
    # ================================================================================================================
    def body(self, frame:tk.Frame):
        """
        :param frame: the frame where you are gonna put stuff
        """

        # ------------------------------------------------------------------------------------------------------------
        # for adding blocks only
        # ------------------------------------------------------------------------------------------------------------
        number_of_blocks_frame = tk.Frame(frame)
        if self.add_edit_type == 'add':
            tk.Label(number_of_blocks_frame, text="Number of Classes:", anchor='e', width=20).pack(side='left', padx=10, pady=5)
            num_entry = entry_float(number_of_blocks_frame,textvariable=self.tk_new_blocks)
            num_entry.pack(side='left', padx=10, pady=5)

        # ------------------------------------------------------------------------------------------------------------
        # duration
        # ------------------------------------------------------------------------------------------------------------
        entry_frame = tk.Frame(frame)
        tk.Label(entry_frame, text="Duration:", anchor='e', width=20).pack(side='left',padx=10,pady=5)
        duration_entry = entry_int(entry_frame, textvariable=self.tk_duration)
        duration_entry.pack(side='left',padx=10, pady=5)

        # ------------------------------------------------------------------------------------------------------------
        # Teacher/Lab Add/Remove
        # ------------------------------------------------------------------------------------------------------------
        teacher_assignments_frame = tk.Frame(frame)
        AddRemoveTk(teacher_assignments_frame, self._non_assigned_teachers, self._assigned_teachers,
                     "Assign Teacher to Class", "Remove Teacher from Class")

        lab_assignments_frame = tk.Frame(frame)
        AddRemoveTk(lab_assignments_frame, self._non_assigned_labs, self._assigned_labs,
                     "Assign Lab to Class","Remove Lab from Class")


        # ------------------------------------------------------------------------------------------------------------
        # layout
        # ------------------------------------------------------------------------------------------------------------
        number_of_blocks_frame.grid(row=0, column=0)
        entry_frame.grid(row=1, column = 0)
        teacher_assignments_frame.grid(row=2,column=0, sticky='nsew', padx=5, pady=5)
        lab_assignments_frame.grid(row=3,column=0, sticky='nsew', padx=5, pady=5)

        return duration_entry

    # ================================================================================================================
    # validate before applying
    # ================================================================================================================
    def validate(self):
        """
        Is the data, as entered by the user, valid?
        :return: True if data is good (changes are applied), false otherwise (nothing happens)
        """
        if not validate_float(self.tk_duration.get(), "Class Duration", "Duration of class must be a valid float!"):
            return False
        if not validate_int(self.tk_new_blocks.get(),"Number of Classes", "Number of classes must be a valid int!"):
            return False
        return True

    # ================================================================================================================
    # apply changes
    # ================================================================================================================
    def apply(self):
        """apply the changes and close the dialog"""
        duration = float(self.tk_duration.get())
        number_new_blocks = int(float(self.tk_new_blocks.get()))
        self._apply_changes(number_new_blocks, duration, self._assigned_teachers, self._assigned_labs )






"""
# ============================================================================
# MAIN PAGE - Gui entry point to the Scheduler Program
#
# - inherits from class MainPageTk
#
# EVENT HANDLERS - to set handlers, use 'set_main_page_event_handler'
#   file_open
#   file_new
#   file_open_previous
#   file_exit
#   semester changed
# ============================================================================
"""
import os.path
from PIL import Image, ImageTk

from typing import Callable, Optional, Literal, get_args

from ..Utilities.Preferences import Preferences

from .main_pages_tk import MainPageBaseTk
import tkinter as tk

from ..gui_generics.read_only_text_tk import ReadOnlyTextTk

BUTTON_WIDTH = 50
MAX_LEN_OF_DISPLAYED_FILENAME = 60

# ============================================================================
# event handlers
# ============================================================================

EVENT_HANDLER_NAMES = Literal[
    "file_open",
    "file_new",
    "file_open_previous",
    "file_exit",
    "semester_change"
]
MAIN_PAGE_EVENT_HANDLERS: dict[EVENT_HANDLER_NAMES, Callable[[], None]] = {}
for event_name in get_args(EVENT_HANDLER_NAMES):
    MAIN_PAGE_EVENT_HANDLERS[event_name] = lambda: print(f"'selected. {event_name}")


def set_main_page_event_handler(name: EVENT_HANDLER_NAMES, handler: Callable[[], None]):
    MAIN_PAGE_EVENT_HANDLERS[name] = handler


# ============================================================================
# class: SchedulerTk
# ============================================================================
class SchedulerTk(MainPageBaseTk):
    """
    GUI_Pages Code for the main Scheduling application window. Inherits from MainPageBaseTk.
    """

    # ========================================================================
    # constructor
    # ========================================================================
    def __init__(self, title: str, preferences: Preferences, bin_dir: str):
        """
        create object - starts with welcome page displayed
        :param title: Sir, Ma'am, Her Royal Highness
        :param preferences: what d'ya want, what d'ya really really want?
        :param bin_dir: the directory that the logo for this app is located
        """
        self.preferences: Preferences = preferences
        self.previous_file_button: Optional[tk.Button] = None
        self._schedule_filename = ""

        super().__init__(title, preferences)
        MainPageBaseTk.bin_dir = bin_dir
        self._current_semester: tk.StringVar = tk.StringVar()  # bound to semester_frame radio buttons
        self._previous_file: tk.StringVar = tk.StringVar(value="None")  # bound to previous_file_button

        # create icon for program
        if title != "Testing":
            ico = Image.open(f"{bin_dir}/schedule_ico.png")
            photo = ImageTk.PhotoImage(ico)
            self.mw.wm_iconphoto(True, photo)

    @property
    def current_semester(self):
        return self._current_semester.get()

    @current_semester.setter
    def current_semester(self, value):
        self._current_semester.set(value)

    @property
    def previous_file(self):
        return self._previous_file.get()

    @previous_file.setter
    def previous_file(self, value):
        if value is not None and value != "":
            basename = os.path.basename(str(value))
            if len(basename) > MAX_LEN_OF_DISPLAYED_FILENAME:
                basename = "..." + basename[len(basename) - MAX_LEN_OF_DISPLAYED_FILENAME:]
            self._previous_file.set(basename)
        else:
            self._previous_file.set("None")

    @property
    def schedule_filename(self):
        return self._schedule_filename

    @schedule_filename.setter
    def schedule_filename(self, value: str):
        self._schedule_filename = value
        if value is not None and value != "":
            value = os.path.abspath(value)
            if len(value) > MAX_LEN_OF_DISPLAYED_FILENAME:
                value = "... " + value[len(value) - MAX_LEN_OF_DISPLAYED_FILENAME:]
        if self.preferences.semester() == "winter":
            self.status_bar_winter_file_info = value
        else:
            self.status_bar_fall_file_info = value

    # ========================================================================
    # override exit event
    # ========================================================================
    def _exit_schedule(self, *_):
        self.exit_schedule()

    def exit_schedule(self):
        MAIN_PAGE_EVENT_HANDLERS["file_exit"]()
        super()._exit_schedule()

    # ========================================================================
    # create Welcome page
    # ========================================================================
    def create_welcome_page(self, semester):
        """Creates the very first page that is shown to the user."""
        option_frame = super().create_welcome_page_base(os.path.join(self.bin_dir, 'schedule_logo.png'))

        # --------------------------------------------------------------
        # which semester?
        # --------------------------------------------------------------
        self._current_semester.set(str(semester))
        semester_frame = tk.Frame(option_frame, background=self.colours.DataBackground)
        semester_frame.pack(side="top", fill="y", expand=1)
        radio_fall = tk.Radiobutton(semester_frame, text='Fall',
                                 background=self.colours.DataBackground, variable=self._current_semester,
                                 font=self.fonts.big,
                                 value='fall')
        radio_winter = tk.Radiobutton(semester_frame, text='Winter',
                                   background=self.colours.DataBackground, variable=self._current_semester,
                                   font=self.fonts.big,
                                   value='winter')

        radio_fall.grid(row=0, column=0, sticky="nsew", padx=25, pady=15)
        radio_winter.grid(row=0, column=1, sticky="nsew", padx=25, pady=15)

        # place a watch on changing bound variables
        self._current_semester.trace('w', MAIN_PAGE_EVENT_HANDLERS["semester_change"])

        def _previous_file_change_event(*_):
            try:
                if self._previous_file != "None":
                    self.previous_file_button.configure(state='normal')
                else:
                    self.previous_file_button.configure(state='disabled')
            except tk.TclError:
                pass
        self._previous_file.trace('w', _previous_file_change_event)

        # --------------------------------------------------------------
        # open previous schedule file buttons
        # -------------------------------------------------------------
        self.previous_file_button = tk.Button(
            option_frame,
            justify="right",
            font=self.fonts.big,
            borderwidth=2,
            background=self.colours.ButtonBackground,
            foreground=self.colours.ButtonForeground,
            command=MAIN_PAGE_EVENT_HANDLERS["file_open_previous"],
            width=BUTTON_WIDTH,
            height=3,
            textvariable=self._previous_file
        )
        self.previous_file_button.pack(side="top", fill="y", expand=0, padx=5, pady=5)

        # --------------------------------------------------------------
        # create new schedule file option
        # --------------------------------------------------------------
        tk.Button(
            option_frame,
            text="Create NEW Schedule File",
            font=self.fonts.big,
            borderwidth=2,
            background=self.colours.ButtonBackground,
            foreground=self.colours.ButtonForeground,
            command=MAIN_PAGE_EVENT_HANDLERS["file_new"],
            width=BUTTON_WIDTH,
            height=3
        ).pack(side="top", fill="y", expand=0, padx=5, pady=5)

        # --------------------------------------------------------------
        # open schedule file option
        # --------------------------------------------------------------
        tk.Button(
            option_frame,
            text="Browse for Schedule File",
            font=self.fonts.big,
            borderwidth=2,
            background=self.colours.ButtonBackground,
            foreground=self.colours.ButtonForeground,
            command=MAIN_PAGE_EVENT_HANDLERS["file_open"],
            width=BUTTON_WIDTH,
            height=3
        ).pack(side="top", fill="y", expand=0, padx=5, pady=5)

        # a growable and shrinkable frame which makes resizing look better
        tk.Frame(
            option_frame, background=self.colours.DataBackground
        ).pack(expand=1, fill="both")

        # --------------------------------------------------------------
        # update the current file and/or button
        # --------------------------------------------------------------
        MAIN_PAGE_EVENT_HANDLERS["semester_change"]()

    # ========================================================================
    # show a text box
    # ========================================================================
    def show_text(self, title, text:list[str]):
        tl = tk.Toplevel(self.mw)
        tl.protocol('WM_DELETE_WINDOW', tl.destroy)
        tl.title(title)
        ro = ReadOnlyTextTk(tl, text, height=20, width=50, scrollbars='e')




PREP_FACTOR                = 0.9
PREP_BONUS_LIMIT           = 3.0
PREP_BONUS_FACTOR          = 0.2  # 1.1 - 0.9
PREP_CRAZY_BONUS_FACTOR    = 1.0

PES_FACTOR                 = 0.04
PES_BONUS_LIMIT            = 415
PES_BONUS_FACTOR           = 0.03

STUDENT_BONUS_LIMIT        = 75
STUDENT_BONUS_FACTOR       = 0.01
STUDENT_CRAZY_BONUS_LIMIT  = 160
STUDENT_CRAZY_BONUS_FACTOR = 0.1

HOURS_FACTOR               = 1.2

CI_FTE_PER_SEMESTER        = 40



from __future__ import annotations
import tkinter as tk
from functools import partial

from ..modified_tk import Scrolled
from ..gui_generics.number_validations import validate_int, entry_int

# =======================================================================================================
# data class for holding info about sections
# =======================================================================================================
class SectionData:
    def __init__(self, section_name, number_of_students, handler):
        self.name = section_name
        self.number_of_students = number_of_students
        self.handler = handler



# =======================================================================================================
# Student Numbers
# =======================================================================================================
class StudentNumbersTk:
    def __init__(self, frame: tk.Frame, data:dict[str, list[SectionData]] = None):
        """
        creates the Panes for each semester
        :param frame:
        :param data: information about the sections. dict[Course Name] = Section Data
        """

        self.frame = frame
        for w in frame.winfo_children():
            w.destroy()

        self.data = data if data is not None else {}
        self._update_handlers = []

        f = Scrolled(
            self.frame,
            "Frame",
            scrollbars='e',
            border=5,
            relief='flat',
        )

        self.pane = f

        # put data into widget
        self.refresh()
        self.pane.bind('<Leave>', func=self.save)

    # -----------------------------------------------------------------------------------------------------------------
    # recreate the gui based on the information in self.data
    # -----------------------------------------------------------------------------------------------------------------
    def refresh(self):
        """
        Refresh the gui
        """
        self._update_handlers.clear()

        if self.pane is None:
            return

        frame = self.pane.widget
        for w in frame.winfo_children():
            w.destroy()

        frame.columnconfigure(1, weight=5)
        frame.columnconfigure(0, weight=1)

        row = 1
        for course_name in self.data.keys():
            tk.Label(frame,
                     text=course_name,
                     anchor='w',
                     width=40
                     ).grid(column=0, row=row, columnspan=2, sticky='nsew', pady=(10,0))
            row += 1

            for section_data in self.data[course_name]:
                tk.Label(frame,
                         width=4,
                         text=section_data.name,
                         anchor='e',
                         ).grid(column=0, row=row, sticky='nsew')
                tk_var = tk.StringVar(value=section_data.number_of_students)
                entry = entry_int(frame, tk_var)
                entry.grid(column=1, row=row, sticky='nw')
                entry.bind("<Return>", lambda e: e.widget.tk_focusNext().focus())
                entry.bind("<FocusIn>", lambda e: self.pane.see(e.widget))

                # create a separate handler for every entry widget, called when
                # this entire widget loses focus (as opposed to every time the data is
                # modified)
                self._update_handlers.append(partial(_update_number, tk_var, section_data.handler))
                row += 1

    # -----------------------------------------------------------------------------------------------------------------
    # called when the user focus for the entry widget is lost
    # -----------------------------------------------------------------------------------------------------------------
    def save(self, *_):
        """Call each save or update handlers.
        Inputs to the handler are
        """
        for handler in self._update_handlers:
            handler()

# =======================================================================================================
# update number
# =======================================================================================================
def _update_number(tk_var: tk.StringVar, handler, *_):
    """
    :param tk_var: the variable holding the data in the entry widget
    :param handler: the presenter defined function to update the section number
    :param _: tk event
    """
    if validate_int(tk_var.get(),"Invalid Int","number of students must be a valid number"):
        handler(int(tk_var.get()))



import pytest
import sys
from os import path

from src.scheduling_and_allocation.model import Lab, TimeSlot, WeekDay


# ============================================================================
# Lab
# ============================================================================
def test_number_getter():
    """Verifies that number getter works as intended."""
    num = "R-101"
    lab = Lab(num)
    assert lab.number == num


def test_descr_getter():
    """Verifies that Lab's name getter works as intended."""
    room = "R-101"
    descr = "Worst place in the world."
    lab = Lab(room, descr)
    assert lab.description == descr


def test_descr_setter():
    """Verifies that the name setter works as intended."""
    lab = Lab()
    descr = "Worst place in the world."
    lab.descr = descr
    assert lab.descr == descr


def test_add_unavailable():
    time_slot1 = TimeSlot(WeekDay.Wednesday, 12.5, 3.5, False)
    time_slot2 = TimeSlot(WeekDay.Thursday, 12.5, 3.5, False)
    lab = Lab()
    lab.add_unavailable_slot(time_slot1)
    lab.add_unavailable_slot(time_slot2)
    assert len(lab.unavailable_slots()) == 2


def test_remove_unavailable_good():
    """Verifies that remove_unavailable_slot() can remove a TimeSlot from Lab based on the received
    TimeSlot ID. """
    day = WeekDay.Monday
    start = 8.5
    dur = 2.0
    lab = Lab()
    t1 = lab.add_unavailable_slot(TimeSlot(day, start, dur))
    t2 = lab.add_unavailable_slot(TimeSlot(day, start, dur))
    lab.remove_unavailable_slot(t1)
    assert len(lab.unavailable_slots()) == 1
    assert lab.unavailable_slots()[0] == t2


def test_unavailable():
    """Verifies that unavailable() returns a list of all unavailable TimeSlots for this Lab."""
    day_1 = WeekDay.Monday
    start_1 = 8.5
    dur_1 = 2.0
    day_2 = WeekDay.Tuesday
    start_2 = 10.0
    dur_2 = 1.5
    lab = Lab()
    lu1 = lab.add_unavailable_slot(TimeSlot(day_1, start_1, dur_1))
    lu2 = lab.add_unavailable_slot(TimeSlot(day_2, start_2, dur_2))
    times = lab.unavailable_slots()
    assert len(times) == 2
    assert lu1 in lab.unavailable_slots()
    assert lu2 in lab.unavailable_slots()


def test_unavailable_no_slots():
    """Verifies that unavailable() returns an empty tuple if no TimeSlots have been assigned to
    this Lab. """
    lab = Lab()
    slots = lab.unavailable_slots()
    assert type(slots) is tuple and len(slots) == 0


def test_string_representation_full():
    """Verifies that the string representation returns a string containing the Lab's number and
    title attributes. """
    num = "R-101"
    desc = "Worst place in the world."
    lab = Lab(num, desc)
    to_string = str(lab)
    assert f"{num}: {desc}" in to_string


def test_string_representation_short():
    """Verifies that string representation returns a string containing only the Lab's number if
    it lacks a title attribute. """
    num = "R-101"
    lab = Lab(num)
    desc = str(lab)
    assert num == desc




import os
import sys

from src.scheduling_and_allocation.gui_pages.allocation_manager_tk import AllocationManagerTk
from src.scheduling_and_allocation.model import SemesterType

bin_dir: str = os.path.dirname(os.path.realpath(__file__))
print (os.path.join(bin_dir, "../../"))
logo_dir = os.path.join(bin_dir, "../../src/scheduling_and_allocation")
c = AllocationManagerTk("Hi",bin_dir = logo_dir)
c.create_welcome_page([SemesterType.winter, SemesterType.fall, SemesterType.summer])
c.start_event_loop()


"""
# ============================================================================
# A view of the class times for a specific resource
#
# Events triggered by ViewDynamicTk
#   tree_popup_menu(gui_id)
#   draw_blocks()
#   on_closing(_*)
#   open_companion_view(gui_id)
#   gui_block_is_moving(gui_id,  gui_block_day, gui_block_start_time)
#   gui_block_has_dropped(gui_id)
#
# From pop-up menu
#   is_movable(block,gui_id)
#   move_resource(block,resource)
#
# Methods called from ViewsController
#   close()
#   draw()
#   is_block_in_view(moved_block)
#   move_gui_block_to(moved_block, day, start_time)
#   refresh_block_colours()
#
# ============================================================================
"""
from __future__ import annotations
import re
from functools import partial
from typing import TYPE_CHECKING, Optional, Callable

from ..gui_generics.menu_and_toolbars import MenuItem, MenuType
from ..Utilities.id_generator import IdGenerator
from ..gui_pages.view_dynamic_tk import ViewDynamicTk
from ..model import Block, Teacher, Stream, Lab, Schedule
from ..model.enums import ResourceType
if TYPE_CHECKING:
    from ..presenter.views_controller import ViewsController

RESOURCE = Lab | Stream | Teacher

# =====================================================================================================================
# use an generator to always guarantee that the new id will be unique
# =====================================================================================================================
_gui_block_ids = IdGenerator()

# =====================================================================================================================
# View
# =====================================================================================================================
class View:
    """View - describes the visual representation of a Schedule."""
    def __init__(self, views_controller: ViewsController, frame, schedule: Schedule,
                 resource: RESOURCE, dirty_flag_method:Callable, gui:ViewDynamicTk=None,
                 ):
        """
        :param views_controller: somebody needs to be in control
        :param frame: container
        :param schedule:
        :param resource:
        :param gui: doesn't need to be set except when setting up for a test suite
        """
        self.views_controller = views_controller
        self.frame = frame
        self.resource = resource
        self.schedule = schedule
        self.gui_blocks: dict[str, Block] = {}
        self.dirty_flag_method = dirty_flag_method

        # get the resource type from the resource object, and all its blocks
        resource_type = ResourceType.none
        if isinstance(resource, Teacher):
            resource_type = ResourceType.teacher
        elif isinstance(resource, Stream):
            resource_type = ResourceType.stream
        elif isinstance(resource, Lab):
            resource_type = ResourceType.lab
        self.resource_type = resource_type

        self.blocks: list[Block] = []

        # create the gui representation of the view
        if gui is None:
            self.gui: ViewDynamicTk = ViewDynamicTk(self.frame, str(resource), resource_type= self.resource_type)
        else:
            self.gui = gui

        # Need to set handlers here, instead on constructor, so that it is easier for testing
        self.gui.get_popup_menu_handler = self.tree_popup_menu
        self.gui.refresh_blocks_handler = self.draw_blocks
        self.gui.on_closing_handler = self.on_closing
        self.gui.double_click_block_handler = self.open_companion_view
        self.gui.gui_block_is_moving_handler = self.gui_block_is_moving
        self.gui.gui_block_has_dropped_handler = self.gui_block_has_dropped
        self.gui.undo_handler = self.views_controller.undo
        self.gui.redo_handler = self.views_controller.redo

        self._block_original_start_time: Optional[float]= None
        self._block_original_day: Optional[float] = None
        self.gui.draw()

    # ----------------------------------------------------------------------------------------------------------------
    # get popup menu (get_popup_menu_handler)
    # ----------------------------------------------------------------------------------------------------------------
    def toggle_is_movable(self, block: Block, gui_id:str):
        """
        toggle movability for block, and inform View Controller
        :param block: the block that needs its 'movable' option toggled
        """
        block.movable = not block.movable
        self.gui.modify_movable(gui_id, block.movable)

        # very important, let the gui controller _know_ that the gui block has been modified
        self.views_controller.notify_block_movable_toggled( block)
        self.dirty_flag_method(True)

        # new action by user, so clear all 'redo'
        self.views_controller.remove_all_redoes()

        # save the action so that it can be undone
        self.views_controller.save_action_block_movable_toggled(block, not block.movable)


    # ----------------------------------------------------------------------------------------------------------------
    # get all the blocks for this resource/schedule (refresh_blocks_handler)
    # ----------------------------------------------------------------------------------------------------------------
    def draw_blocks(self):
        """
        draw blocks onto Canvas
        """

        # get all the blocks for this resource/schedule
        blocks = []
        match self.resource_type:
            case ResourceType.teacher:
                blocks = self.schedule.get_blocks_for_teacher(self.resource)
            case ResourceType.stream:
                blocks = self.schedule.get_blocks_for_stream(self.resource)
            case ResourceType.lab:
                blocks = self.schedule.get_blocks_in_lab(self.resource)

        # create all the gui blocks and draw them
        self.gui_blocks.clear()

        for block in blocks:

            gui_block_id: int = _gui_block_ids.get_new_id()
            gui_tag = f"gui_tag_{gui_block_id}"
            self.gui_blocks[gui_tag] = block

            text = self.get_block_text(block, scale = self.gui.get_scale(),resource_type = self.resource_type)
            day, start_time, duration = self._block_to_floats(block)

            self.gui.draw_block(resource_type=self.resource_type,
                                day=day, start_time=start_time, duration=duration,
                                text=text,
                                gui_block_id=gui_tag,
                                movable=block.movable)
            self.gui.colour_block(gui_tag, self.resource_type, block.movable, conflict=block.conflict)

    # ----------------------------------------------------------------------------------------------------------------
    # is block in this view?
    # ----------------------------------------------------------------------------------------------------------------
    def is_block_in_view(self, block):
        return block in self.gui_blocks.values()

    # ----------------------------------------------------------------------------------------------------------------
    # important tidy-up stuff (on_closing_handler)
    # ----------------------------------------------------------------------------------------------------------------
    def on_closing(self, *_):
        """When the view closes, clean up"""
        self.views_controller.view_is_closing(self.resource)

    def close(self):
        self.gui.kill()

    # ----------------------------------------------------------------------------------------------------------------
    # open companion view (double_click_handler)
    # ----------------------------------------------------------------------------------------------------------------
    def open_companion_view(self, gui_tag: str):
        """
        Based on the resource_type of this view, will open another view which has this Block.
        :param gui_tag:
        """
        block = self.gui_blocks[gui_tag]
        self.views_controller.open_companion_view(block)

    # ----------------------------------------------------------------------------------------------------------------
    # gui_block_is_moving (gui_block_is_moving_handler)
    # ----------------------------------------------------------------------------------------------------------------
    def gui_block_is_moving(self, gui_id: str,  gui_block_day: float, gui_block_start_time: float):
        """
        Update block and conflict information as the user is dragging the gui block
        :param gui_id: the id of the gui representation of the block
        :param gui_block_day: the position of the gui block
        :param gui_block_start_time: the start time of the gui block
        """

        block: Block = self.gui_blocks.get(gui_id,None)
        if block is None:
            return

        # capture the info about the block before it starts moving
        if self._block_original_start_time is None and self._block_original_day is None:
            self._block_original_start_time = block.start
            self._block_original_day = block.day.value

        # update the block by snapping to time and day, although it doesn't update the gui block
        block.start = gui_block_start_time
        block.snap_to_day(gui_block_day)
        block.snap_to_time()
        self.schedule.calculate_conflicts()
        self.refresh_block_colours()
        #self.gui.colour_block(gui_id, self.resource_type, is_movable=block.movable, conflict = block.conflict)

        # very important, let the gui controller _know_ that the gui block has been moved
        self.views_controller.notify_block_move(self.resource.number, block, gui_block_day, gui_block_start_time)

        # don't set dirty flag until block has dropped!

    # ----------------------------------------------------------------------------------------------------------------
    # gui_block_has_dropped (gui_block_has_dropped_handler)
    # ----------------------------------------------------------------------------------------------------------------
    def gui_block_has_dropped(self, gui_id: str):
        """
        User has stopped dragging the gui block, so adjust gui block to block's new location and
        inform View Controller
        :param gui_id: the id of the gui representation of the block
        """
        block: Block = self.gui_blocks.get(gui_id,None)
        if block is None:
            return

        # set the gui block coordinates to match the block
        # (which is constantly being snapped to grid during the move process)
        self.gui.move_gui_block(gui_id, block.day.value,block.start)

        # has the block actually moved?
        if self._block_original_start_time is None or self._block_original_day is None:
            return

        if (self._block_original_start_time == block.start and
            self._block_original_day == block.day):
            return

        # very important, let the gui controller _know_ that the gui block has been moved
        self.views_controller.notify_block_move(self.resource.number, block, block.day.value,
                                                block.start)

        # new action by user, so clear all 'redo'
        self.views_controller.remove_all_redoes()

        # save the action so that it can be undone
        self.views_controller.save_action_block_move(block, self._block_original_day, block.day.value,
                                                     self._block_original_start_time, block.start)

        # reset the 'start move' info
        self._block_original_start_time = None
        self._block_original_day = None

        self.dirty_flag_method(True)

    # ----------------------------------------------------------------------------------------------------------------
    # get popup menu (get_popup_menu_handler)
    # ----------------------------------------------------------------------------------------------------------------
    def tree_popup_menu(self, gui_id) -> Optional[list[MenuItem]]:
        """
        Create a pop-up menu that is unique to this gui_id
        :param gui_id:
        """
        block = self.gui_blocks.get(gui_id, None)

        if block is None:
            return

        # get the resources (note we cannot swap blocks between streams)
        resources = set()
        match self.resource_type:
            case ResourceType.teacher:
                resources = set(self.schedule.teachers()) - set(block.teachers())
            case ResourceType.lab:
                resources = set(self.schedule.labs()) - set(block.labs())

        # create the menu
        toggle_str = "Unset 'movable' option" if block.movable else "Set 'movable' option"
        menu_list: list[MenuItem] = [MenuItem(menu_type=MenuType.Command, label=toggle_str,
                                              command=partial(self.toggle_is_movable, block, gui_id))]

        if len(resources):
            menu_list.append(MenuItem(menu_type=MenuType.Separator))

        resource_list = []
        for resource in sorted(resources):
            resource_list.append(MenuItem(menu_type=MenuType.Command, label=f"{resource}",
                                      command=partial(self.move_resource, block, resource)
                                      ))
        menu_list.append(MenuItem(menu_type=MenuType.Cascade, label=f"Move class to {self.resource_type.name} ...",
                         children=resource_list))

        return menu_list

    # ----------------------------------------------------------------------------------------------------------------
    # modify the resource (from pop-up menu)
    # ----------------------------------------------------------------------------------------------------------------
    def move_resource(self, block: Block, resource: RESOURCE):
        """
        move resource (ask Views Controller to do it for us)
        :param block:
        :param resource:
        """

        # let gui controller handle this because it involves more than one resource
        self.views_controller.notify_move_block_to_resource(resource_type = self.resource_type,
                                                            block=block,
                                                            from_resource= self.resource,
                                                            to_resource= resource)

        # new action by user, so clear all 'redo'
        self.views_controller.remove_all_redoes()

        # save the action so that it can be undone
        self.views_controller.save_action_block_resource_changed(resource_type = self.resource_type,
                                                                 block=block,
                                                                 from_resource= self.resource,
                                                                 to_resource= resource)

    # ----------------------------------------------------------------------------------------------------------------
    # move_gui_block_to (used by Views Controller)
    # ----------------------------------------------------------------------------------------------------------------
    def move_gui_block_to(self, block: Block, day: float, start_time: float):
        """
        move the gui block (associated with block) to a new position (used by Views Controller)
        :param block: block that is associated with the gui image
        :param day: the day that the **gui block** needs to be moved to (not necessarily the same as the block)
        :param start_time: the time that the **gui block** needs to be moved to (not necessarily the same as the block)
        """
        gui_id = self._get_gui_id_from_block(block.number)
        self.gui.move_gui_block(gui_id, day, start_time)


    # ----------------------------------------------------------------------------------------------------------------
    # refresh block colour (used by Views Controller & View)
    # ----------------------------------------------------------------------------------------------------------------
    def refresh_block_colours(self):
        """Go through all blocks, and adjust colours as required"""

        for gui_tag, block in self.gui_blocks.items():
            self.gui.colour_block(gui_tag, self.resource_type, is_movable=block.movable, conflict = block.conflict)

    # ----------------------------------------------------------------------------------------------------------------
    # get block text
    # ----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def get_block_text( block: Block, resource_type, scale=1):
        """
        :param block: The block object
        :param scale: the scale of the view
        :param resource_type: Teacher | Stream | Lab
        """

        # course & section & streams
        course_number = ""
        section_number = ""
        if block.section:
            course_number = block.section.course.number if scale > 0.5 else re.split("[-*]", block.section.course.number)
            section_number = block.section.title

        # labs
        lab_numbers = ", ".join(l.number for l in block.labs())
        lab_numbers = lab_numbers if scale > .75 else lab_numbers[0:16] + "..."
        lab_numbers = lab_numbers if scale > .50 else lab_numbers[0:7] + "..."
        if resource_type == ResourceType.lab and (scale <= .75 or block.duration <= 1):
                lab_numbers = ""
        lab_numbers = "" if resource_type == ResourceType.lab and block.duration <= 1 else lab_numbers

        # streams
        stream_numbers = ",".join(s.number for s in block.streams())
        stream_numbers = stream_numbers if scale > .75 else stream_numbers[0:11] + "..."
        stream_numbers = stream_numbers if scale > .50 else stream_numbers[0:7] + "..."
        if resource_type == ResourceType.stream and (scale <= .75 or block.duration <= 1):
                stream_numbers = ""

        # teachers
        teachers_name = ""
        if len(block.teachers()) <= 2:
            for t in block.teachers():
                if len(str(t)) > 12:
                    teachers_name = teachers_name + f"{t.firstname} {t.lastname[0:1]}.\n"
                else:
                    teachers_name = teachers_name + f"{str(t)}\n"
        teachers_name = teachers_name.rstrip()
        teachers_name = teachers_name if scale > .75 else teachers_name[0:11] + "..."
        teachers_name = teachers_name if scale > .50 else teachers_name[0:7] + "..."
        if resource_type == ResourceType.teacher:
            if scale < .75 or block.duration <= 1:
                teachers_name = ""

        # define what to display
        block_text = f"{course_number}\n{section_number}\n"
        block_text = block_text + teachers_name + "\n" if teachers_name else block_text
        block_text = block_text + lab_numbers + "\n" if lab_numbers else block_text
        block_text = block_text + stream_numbers + "\n" if stream_numbers else block_text

        block_text = block_text.rstrip()
        return block_text

    # ----------------------------------------------------------------------------------------------------------------
    # draw - called by Views Controller
    # ----------------------------------------------------------------------------------------------------------------
    def draw(self):
        self.gui.draw()

    # ----------------------------------------------------------------------------------------------------------------
    # block to floats
    # ----------------------------------------------------------------------------------------------------------------
    @staticmethod
    def _block_to_floats( block) -> tuple[int,float,float]:
        """take a block and return number representations of its properties"""
        return block.day.value, block.start, block.duration

    # ----------------------------------------------------------------------------------------------------------------
    # get gui id from block
    # ----------------------------------------------------------------------------------------------------------------
    def _get_gui_id_from_block(self, block_number: str) -> Optional[str]:
        for g,b in self.gui_blocks.items():
            if b.number == block_number:
                return g
        return None




"""
# ============================================================================
# Gathers info for main menus and toolbars for scheduler, and gives them default commands
# - to set the commands, use methods 'set_menu_event_handler'
#
# NOTE: this does not actually create the gui menu, it just gathers the info
# ============================================================================
"""
from __future__ import annotations

from typing import Callable, Literal, get_args
from ..gui_generics.menu_and_toolbars import MenuItem, MenuType, ToolbarItem

MAIN_MENU_EVENT_HANDLER_NAMES = Literal[
    "file_new",
    "file_open",
    "file_save",
    "file_save_as",
    "file_exit",
    "print_pdf_teacher",
    "print_pdf_lab",
    "print_pdf_streams",
    "print_text",
    "print_latex_teacher",
    "print_latex_lab",
    "print_latex_streams",
    "print_registrar",
    "validate",
]



MAIN_MENU_EVENT_HANDLERS: dict[MAIN_MENU_EVENT_HANDLER_NAMES, Callable[[], None]] = {}
for event_name in get_args(MAIN_MENU_EVENT_HANDLER_NAMES):
    MAIN_MENU_EVENT_HANDLERS[event_name] = lambda: print(f"'{event_name}' selected. ")


def set_menu_event_handler(name: MAIN_MENU_EVENT_HANDLER_NAMES, handler: Callable[[], None]):
    MAIN_MENU_EVENT_HANDLERS[name] = handler


def main_menu() -> tuple[list[str], dict[str, ToolbarItem], list[MenuItem]]:
    menu = list()
    # NOTE:  if these events are bound, the first parameter passed will be the event.
    #        - we want to ignore these events, because the event handlers don't need this info
    # -----------------------------------------------------------------------------------------
    # File menu
    # -----------------------------------------------------------------------------------------
    file_menu = MenuItem(name='file', menu_type=MenuType.Cascade, label='File')
    file_menu.add_child(MenuItem(name='new', menu_type=MenuType.Command,
                                 label='New', accelerator='Control-n',
                                 command=lambda *_: MAIN_MENU_EVENT_HANDLERS["file_new"]()
                                 )
                        )
    file_menu.add_child(MenuItem(name='open', menu_type=MenuType.Command,
                                 label='Open', accelerator='Control-o',
                                 command=lambda *_: MAIN_MENU_EVENT_HANDLERS["file_open"]()
                                 )
                        )

    file_menu.add_child(MenuItem(menu_type=MenuType.Separator))

    file_menu.add_child(MenuItem(name='save', menu_type=MenuType.Command,
                                 label='Save',
                                 accelerator='Control-s',
                                 underline=False,
                                 command=lambda *_: MAIN_MENU_EVENT_HANDLERS["file_save"]()
                                 )
                        )
    file_menu.add_child(MenuItem(name='save_as', menu_type=MenuType.Command,
                                 label='Save As',
                                 command=lambda *_: MAIN_MENU_EVENT_HANDLERS["file_save_as"]()
                                 )
                        )

    file_menu.add_child(MenuItem(menu_type=MenuType.Separator))

    file_menu.add_child(MenuItem(menu_type=MenuType.Command,
                                 label='Validate',
                                 command=lambda *_: MAIN_MENU_EVENT_HANDLERS["validate"]()
                                 )
                        )
    file_menu.add_child(MenuItem(menu_type=MenuType.Command,
                                 label='Exit',
                                 accelerator='Control-e',
                                 command=lambda *_: MAIN_MENU_EVENT_HANDLERS["file_exit"]()
                                 )
                        )

    # -----------------------------------------------------------------------------------------
    # Print menu
    # -----------------------------------------------------------------------------------------
    print_menu = MenuItem(name='print', menu_type=MenuType.Cascade, label='Print')
    pdf_menu = MenuItem(menu_type=MenuType.Cascade, label='PDF')
    latex_menu = MenuItem(menu_type=MenuType.Cascade, label='Latex')

    print_menu.add_child(MenuItem(menu_type=MenuType.Command,
                                  label="Registrar's Output",
                                  command = lambda *_: MAIN_MENU_EVENT_HANDLERS["print_registrar"]()
                                  )
                         )

    print_menu.add_child(pdf_menu)
    print_menu.add_child(latex_menu)

    # pdf sub-menu
    pdf_menu.add_child(MenuItem(menu_type=MenuType.Command,
                                label='Teacher Schedules',
                                command=lambda *_: MAIN_MENU_EVENT_HANDLERS["print_pdf_teacher"]()
                                )
                       )
    pdf_menu.add_child(MenuItem(menu_type=MenuType.Command,
                                label='Lab Schedules',
                                command=lambda *_: MAIN_MENU_EVENT_HANDLERS["print_pdf_lab"]()
                                )
                       )
    pdf_menu.add_child(MenuItem(menu_type=MenuType.Command,
                                label='Stream Schedules',
                                command=lambda *_: MAIN_MENU_EVENT_HANDLERS["print_pdf_streams"]()
                                )
                       )

    pdf_menu.add_child(MenuItem(menu_type=MenuType.Separator))

    pdf_menu.add_child(MenuItem(menu_type=MenuType.Command,
                                label='Text Output',
                                command=lambda *_: MAIN_MENU_EVENT_HANDLERS["print_text"]()
                                )
                       )

    # latex sub menu
    latex_menu.add_child(MenuItem(menu_type=MenuType.Command,
                                  label='Teacher Schedules',
                                  command=lambda *_: MAIN_MENU_EVENT_HANDLERS["print_latex_teacher"]()
                                  )
                         )
    latex_menu.add_child(MenuItem(menu_type=MenuType.Command,
                                  label='Lab Schedules',
                                  command=lambda *_: MAIN_MENU_EVENT_HANDLERS["print_latex_lab"]()
                                  )
                         )
    latex_menu.add_child(MenuItem(menu_type=MenuType.Command, label='Stream Schedules',
                                  command=lambda *_: MAIN_MENU_EVENT_HANDLERS["print_latex_streams"]()
                                  )
                         )
    # -----------------------------------------------------------------------------------------
    # Auto Save & Preferences - taken care of if main_tk
    # -----------------------------------------------------------------------------------------

    # -----------------------------------------------------------------------------------------
    # _toolbar
    # -----------------------------------------------------------------------------------------
    toolbar_info = dict()
    toolbar_info['new'] = ToolbarItem(command=lambda *_: MAIN_MENU_EVENT_HANDLERS["file_new"](),
                                      hint='Create new Schedule File')

    toolbar_info['open'] = ToolbarItem(command=lambda *_: MAIN_MENU_EVENT_HANDLERS["file_open"](),
                                       hint='Open Schedule File')

    toolbar_info['save'] = ToolbarItem(command=lambda *_: MAIN_MENU_EVENT_HANDLERS["file_save"](),
                                       hint='Save Schedule File')
    toolbar_order = ['new', 'open', '', 'save']

    # return list of top level menu items
    menu.append(file_menu)
    menu.append(print_menu)

    return toolbar_order, toolbar_info, menu



"""Enums for global use"""
from __future__ import annotations
from enum import Enum, Flag
from typing import TYPE_CHECKING

''' Not all the classes are enums, but this file contains data that is needed throughout the
program'''


class ExtendedEnum(Enum):
    @classmethod
    def values(cls):
        return list(map(lambda c: c.value, cls))

    @classmethod
    def names(cls):
        return list(map(lambda c: c.name, cls))

    @classmethod
    def validate(cls, user_input):
        # want to allow user to enter a value,  but also want to
        # allow it to be an enum, for interface reasons (forces the user to pass
        # the correct resource_type)
        if isinstance(user_input, cls):
            return user_input
        elif user_input in cls.names():
            return cls[user_input]
        elif user_input not in cls.values():
            raise ValueError(f"Error: input <{user_input}> is invalid")
        return user_input


class WeekDay(Enum):
    Sunday = 0
    Monday = 1
    Tuesday = 2
    Wednesday = 3
    Thursday = 4
    Friday = 5
    Saturday = 6

    def __lt__(self, other):
        return self.value < other.value

    @staticmethod
    def get_from_string(value) -> WeekDay:
        value = value[0:3].lower()
        return {"mon": WeekDay.Monday,
                "tue": WeekDay.Tuesday,
                "wed": WeekDay.Wednesday,
                "thu": WeekDay.Thursday,
                "fri": WeekDay.Friday,
                "sat": WeekDay.Saturday,
                "sun": WeekDay.Sunday}.get(value, None)


class SemesterType(ExtendedEnum):
    fall = 1
    winter = 2
    summer = 3
    any = 4



class ResourceType(ExtendedEnum):
    lab = "Lab"
    teacher = "Teacher"
    stream = "Stream"
    none = "None"

# =============================================================================
# ConflictType Enum
# =============================================================================
class ConflictType(Flag):
    """All types of Conflict types... note it is not an ENUM, but a FLAG
    which allows `or` and `len`, and `__contains__` built in"""
    NONE = 0
    TIME = 1
    LUNCH = 2
    MINIMUM_DAYS = 4
    AVAILABILITY = 8
    TIME_TEACHER = 16
    TIME_LAB = 32
    TIME_STREAM = 64

    @classmethod
    def colours(cls):
        """return a dictionary of colours to go with each conflict type"""
        return {
            cls.TIME: "indianred3",
            cls.LUNCH: "tan4",
            cls.MINIMUM_DAYS: "lightgoldenrod1",
            cls.AVAILABILITY: "mediumvioletred",
            cls.TIME_TEACHER: "red2",
            cls.TIME_LAB: "red2",
            cls.TIME_STREAM: "red2",
        }

    @classmethod
    def descriptions(cls) -> dict[ConflictType, str]:
        """returns a dictionary of descriptions to go with each conflict type"""
        return {
            cls.NONE: "no conflict",
            cls.TIME: "indirect time overlap",
            cls.LUNCH: "no lunch time",
            cls.MINIMUM_DAYS: "too few days",
            cls.TIME_TEACHER: "time overlap",
            cls.TIME_LAB: "time overlap",
            cls.TIME_STREAM: "time overlap",
            cls.AVAILABILITY: "not available",
        }

    def description(self) -> str:
        """ Returns the title of the provided conflict resource_type """
        return ConflictType.descriptions()[self]

    def is_conflicted(self) -> bool:
        """any conflicts?"""
        return len(self) != 0

    def is_time(self) -> bool:
        """does the conflict number include a time conflict?"""
        return ConflictType.TIME in self

    def is_time_lab(self) -> bool:
        """does the conflict number include a time - lab conflict?"""
        return ConflictType.TIME_LAB in self

    def is_time_teacher(self) -> bool:
        """does the conflict number include a time - teacher conflict?"""
        return ConflictType.TIME_TEACHER in self

    def is_time_stream(self) -> bool:
        """does the conflict number include a time - stream conflict?"""
        return ConflictType.TIME_STREAM in self

    def is_time_lunch(self) -> bool:
        """does the conflict number include a lunch conflict?"""
        return ConflictType.LUNCH in self

    def is_minimum_days(self) -> bool:
        """does the conflict number include a minimum days conflict?"""
        return ConflictType.MINIMUM_DAYS in self

    def is_availability(self) -> bool:
        """does the conflict number include a minimum days conflict?"""
        return ConflictType.AVAILABILITY in self

    def most_severe(self, view_type: ResourceType = ResourceType.none) -> ConflictType:
        """
        Identify the most severe conflict resource_type in a list of conflicts defined by
        the conflict number
        :param view_type: -> defines the user's current view. ResourceType (enum)
        """
        severest = ConflictType.NONE
        sorted_conflicts: list[ConflictType] = ORDER_OF_SEVERITY.copy()

        # based on the view, get_by_id the string version of the enum ConflictType
        if view_type is not None:
            conflict_key: str = f"time_{view_type.name}".upper()

            # if conflict key is a valid enum for ConflictType, insert this into
            # the sorted conflicts list
            if conflict_key in ConflictType.__members__:
                sorted_conflicts.insert(0, ConflictType[conflict_key])

        # look for the first match where the bit in conflict_number matches a possible
        # conflict resource_type
        for conflict in sorted_conflicts:
            if conflict in self:
                severest = conflict
                break

        # it doesn't make sense for labs or streams to have LUNCH, AVAILABILITY, or MINIMUM_DAYS
        # as valid conflicts
        if view_type == ResourceType.lab or view_type == ResourceType.stream:
            if severest == ConflictType.LUNCH or severest == ConflictType.AVAILABILITY or severest == ConflictType.MINIMUM_DAYS:
                severest = ConflictType.NONE


        return severest

ORDER_OF_SEVERITY = [ConflictType.TIME, ConflictType.LUNCH, ConflictType.MINIMUM_DAYS,
                     ConflictType.AVAILABILITY]




from .block import Block
from .enums import ConflictType
from .time_slot import TimeSlot
from .enums import WeekDay, SemesterType, ResourceType
from .lab import Lab
from .teacher import Teacher
from .section import Section
from .stream import Stream
from .course import Course
from .schedule import Schedule
from .exceptions import InvalidSectionNumberForCourseError, InvalidHoursForSectionError, \
    CouldNotReadFileError
from .conflicts import set_block_conflicts, set_lunch_break_conflicts, \
    set_number_of_days_conflict, MAX_HOURS_PER_WEEK, set_availability_hours_conflict
from .time_slot import MINIMUM_DURATION, DEFAULT_DAY, DEFAULT_START, DEFAULT_DURATION, \
    MINUTE_BLOCK_SIZE, MIN_START_TIME, MAX_END_TIME, MAXIMUM_DURATION



from __future__ import annotations

import tkinter as tk
from functools import partial
from typing import Callable, Any, Optional
from .InitGuiFontsAndColours import TkColours



class DragNDropManager:
    _currently_dragging: bool = False  # are we currently dragging something?
    _info_data: dict[Any, Any] = dict()  # dictionary so calling program can exchange data between event handlers
    _drag_indicator: Optional[tk.Toplevel] = None  # a top level window that moves around with the mouse
    _default_cursor: str = ""  # the cursor of associated with the source widget before dragging begins
    _label: tk.Label = None  # the content of the drag_indicator
    _current_target: Optional[tk.Widget] = None  # the current, valid, cb_drop-site widget
    _drag_indicator_text: str = ""  # the text in the label in the drag_indicator

    def __init__(self, colours = TkColours()):
        self.colours = colours

    # ========================================================================
    # add a widget to the 'draggable' list
    # ========================================================================
    def add_draggable(self, source: tk.Widget,
                      on_start: Callable[[tk.Event, dict], str] = lambda e, info_data: str(e.widget),
                      on_drag: Callable[[tk.Event, dict, tk.Widget], bool] = lambda e, info_data, target: bool,
                      on_drop: Callable[[tk.Event, dict, tk.Widget], None] = lambda e, info_data, target: None):
        """
        Add the specified widget to the list of widgets that can be dragged and dropped.

        Details
        -------

        * When the widget is selected, the `on_start` function will be called, where any information required during
          the drag-n-drop phase can be stored in the `info_data` dictionary.  This method should return a string
          which will be used to label the little square that physical indicates where things are being dropped

        * When the widget is being moved about, the `on_drag` method will be called, with the current event,
          the data saved in `info_data`, and a widget that is under the mouse.  Must return true or false to indicate
          if it is a valid drop-site

        * When the widget is dropped, if the mouse is hovering over another widget (target) that has been
          set as a drop-site, the `on_drop` method will be called, again passing in any information stored in
          `info_data`

        * The widget that is being dragged about is available through the `tk.Event` object, (event.widget)

        Parameters
        ----------

        :param source: the widget which will be selected for dropping
        :param on_start:
        :param on_drag:
        :param on_drop:
        :return: Nothing
        """
        source.bind("<ButtonPress-1>", partial(self._on_start, on_start, on_drag, on_drop))

    # ========================================================================
    # Prepare for dragging the widget
    # ========================================================================
    def _on_start(self,
                  cb_start: Callable[[tk.Event, dict], str],
                  cb_drag: Callable[[tk.Event, dict, tk.Event], None],
                  cb_drop: Callable[[tk.Event, dict], None],
                  event: tk.Event):
        """Start the drag'n'drop process"""

        # prevent multiple dragging and dropping
        if self._currently_dragging:
            return

        # call user's 'time_start' method defined for this source widget
        self._info_data.clear()
        self._drag_indicator_text = cb_start(event, self._info_data)
        if not self._drag_indicator_text:
            self._drag_indicator_text = str(event.widget)

        # setup for dragging and dropping object
        self._current_target = None
        self._default_cursor = event.widget.cget("cursor")
        source = event.widget
        source.bind("<B1-Motion>", partial(self._on_drag, cb_drag))
        source.bind("<ButtonRelease-1>", partial(self._on_drop, cb_drop))

    # ========================================================================
    # Drag the widget
    # ========================================================================
    def _on_drag(self, cb_drag: Callable[[tk.Event, dict, tk.Widget], None], event: tk.Event):

        # create a popup window for dragging
        if not self._drag_indicator:
            tl = tk.Toplevel(background = self.colours.DataBackground)
            self._label = tk.Label(tl, text=self._drag_indicator_text, foreground=self.colours.DataForeground)
            self._label.pack(expand=1, fill="both")
            tl.update_idletasks()
            tl.overrideredirect(True)
            tl.transient()
            self._drag_indicator = tl
            event.widget.configure(cursor="hand2")

        self._label.configure(bg=self.colours.DragNDropBackground)
        self._label.configure(fg=self.colours.DragNDropForeground)
        tl = self._drag_indicator

        # move the indicator by specifying the geometry
        tl.geometry(f"100x20+{event.x_root + 5}+{event.y_root - 5}")

        # find the target under the cursor
        target = event.widget.winfo_containing(event.x_root, event.y_root)

        # call user's 'drag' method, which returns 'true' if mouse over valid drop-site
        if cb_drag(event, self._info_data, target):
            self._label.configure(bg=self.colours.DragNDropAllowedBackground)

    # ========================================================================
    # Drop the widget
    # ========================================================================
    def _on_drop(self, cb_drop: Callable[[tk.Event, dict, tk.Widget], None], event: tk.Event):

        # find the target under the cursor
        target = event.widget.winfo_containing(event.x_root, event.y_root)

        # call user's 'drop' method
        cb_drop(event, self._info_data, target)

        # clean up after dropping
        if self._drag_indicator:
            self._drag_indicator.destroy()
        self._drag_indicator = None
        self._currently_dragging = False
        event.widget.configure(cursor=self._default_cursor)
        event.widget.unbind("<B1-Motion>")
        event.widget.unbind("<ButtonRelease-1>")



from functools import partial
import tkinter as tk

from src.scheduling_and_allocation.gui_pages.student_numbers_tk import SectionData, StudentNumbersTk
from src.scheduling_and_allocation.model import Schedule
from src.scheduling_and_allocation.presenter.student_numbers import StudentData

import os

bin_dir: str = os.path.dirname(os.path.realpath(__file__))

file = f"{bin_dir}/../unit_tests_presenter/data_fall.csv"
schedule = Schedule(file)

data: StudentData = StudentData()

def refresh():
    for course in (c for c in schedule.courses() if c.needs_allocation):
        data.add_course(course.title)

        for section in course.sections():
            section_data = SectionData(section.name, section.num_students, partial(data_changed_handler, section))
            data.add_section(course.title, section_data)

    return data

def data_changed_handler(section, number:int):
    section.num_students = number
refresh()


mw = tk.Tk()
frame = tk.Frame(mw)
frame.pack(expand=1, fill='both')

ns = StudentNumbersTk(frame, data.courses)


mw.mainloop()



"""
Provides canvases for drawing views, other than the Tk canvas
"""
from reportlab.lib.pagesizes import LETTER
from reportlab.pdfgen.canvas import Canvas
from datetime import datetime

from ..Utilities import Colour
from ..model import ResourceType

import pathlib
CODE_PATH = pathlib.Path(__file__).parent.resolve()
CWD = pathlib.Path().resolve()

# ====================================================================================================================
# Latex
# ====================================================================================================================
class LatexCanvas:
    """Doesn't 'draw' a canvas so much as uses a template and inserts latex stuff"""

    def __init__(self, title="Title", schedule_name="sub_title", filename=None, directory=None):
        """
        creates the latex file to work with
        :param title: title to be displayed
        :param schedule_name: the filename of the schedule
        :param filename: the file to save the tex file to (defaults to title, in current working directory)
        """
        if directory is None:
            directory = CWD
        if filename is None:
            filename = title.lower()
            filename = filename.replace(" ","_")
            filename = f"{directory}/{filename}.tex"

        self.filename = filename
        template_file = open(f"{CODE_PATH}/view_template.tex","r")
        self.tex = template_file.read()
        template_file.close()
        title = title.replace("_",r"\_")
        schedule_name = schedule_name.replace("_",r"\_")

        self.tex = self.tex.replace("%%USER_NAME%%", str(title))
        self.tex = self.tex.replace("%%EMAIL%%","")
        self.tex = self.tex.replace("%%PHONE_NUMBER%%","")
        self.tex = self.tex.replace("%%FILENAME%%", schedule_name)
        self.tex = self.tex.replace("%%PRINTED_DATE%%",datetime.today().strftime('%Y-%m-%d'))

        self.blocks = []


    def draw_block(self, resource_type: ResourceType, day: int, start_time: float,
                   duration: float, text:str, gui_tag, movable):
        """
        doesn't actually draw a block, but it keeps the tex info required in a list
        :param resource_type:
        :param day:
        :param start_time:
        :param duration:
        :param text:
        :param gui_tag:
        :param movable:
        """
        day_name = ["sunday", "monday", "tuesday", "wednesday", "thursday", "friday", "saturday"]
        text = text.replace("\n",r" \\ ")
        self.blocks.append(
            r"\node[course={"+f"{duration}"+"}{1}] at "+fr"(\{day_name[int(day)]},{start_time}) "+"{"+text+" };"
        )

    def save(self):
        """writes the tex file"""
        self.tex = self.tex.replace("%%CLASS_TIMES%%", "\n".join(self.blocks))
        fh = open(self.filename, "w")
        fh.write(self.tex)
        fh.close()


# ====================================================================================================================
# PDF
# ====================================================================================================================
class PDFCanvas:
    """Provides a canvas for drawing a pdf"""

    def __init__(self, title="Title", schedule_name="sub_title", filename=None, directory=None):
        """
        creates a pdf canvas to draw on
        :param title: title to be displayed
        :param schedule_name: the filename of the schedule
        :param filename: the file to save the pdf file to (defaults to title, in current working directory)
        """

        if directory is None:
            directory = CWD
            
        if filename is None:
            filename = title.lower()
            filename = filename.replace(" ","_")
            filename = f"{directory}/{filename}.pdf"

        self.cn = Canvas(filename, bottomup=0, pagesize=LETTER)
        self.cn.setFont("Helvetica", 14)
        self.cn.drawCentredString(LETTER[0]/2,50, str(title))
        self.cn.setFont("Helvetica", 10)
        self.cn.drawCentredString(LETTER[0]/2,70, str(schedule_name))
        self.cn.setFont("Helvetica", 6)
        self.cn.drawCentredString(LETTER[0]/2,80, f"Printed: {datetime.today().strftime('%Y-%m-%d')}")

        self.cn.setFont("Helvetica", 8)
        self.cn.setLineWidth(0.5)
        self.cn.translate(30,100)

    def config(self,*args, **kwargs):
        pass

    def create_line(self, x1:float, y1:float, x2:float, y2:float, fill:str="grey", dash="", tags: str|tuple=""):
        """draws a line"""
        if dash == "-":
            self.cn.setDash([6,3])
        elif dash == ".":
            self.cn.setDash([1,2])
        r,g,b = Colour.rgb(fill)
        self.cn.setFillColorRGB(r,g,b)
        self.cn.line(x1,y1,x2,y2)

    def addtag_withtag(self,*args,**kwargs):
        """not needed for this canvas"""
        pass

    def create_text(self, x:float, y:float, text:str="", fill:str='black', tags:str|tuple="")->int:
        """
        draws the text, centered on x/y
        :param x:
        :param y:
        :param text:
        :param fill:
        :param tags: not used
        """
        r,g,b = Colour.rgb("black")
        self.cn.setFillColorRGB(r,g,b)
        lines = text.split("\n")
        baseline = 10
        y = y - (len(lines)-1)*baseline/2
        for t in text.split("\n"):
            self.cn.drawCentredString(x,y,t)
            y = y+10
        return 1

    def create_rectangle(self, coords:tuple[float,float,float,float], fill:str='grey', outline:str='grey',
                                        tags:str|tuple="")->int:
        """
        creates a rectangle with bounding coordinates
        :param coords:
        :param fill:
        :param outline:
        :param tags: (not used)
        """

        r,g,b = Colour.rgb(outline)
        self.cn.setStrokeColorRGB(r,g,b)
        r,g,b = Colour.rgb(fill)
        self.cn.setFillColorRGB(r,g,b)
        x1,y1,x2,y2 = coords
        self.cn.rect(min(x1,x2),min(y1,y2),abs(x2-x1), abs(y2-y1), fill=1)

        return 1

    def save(self):
        """save the pdf"""
        self.cn.save()




from tkinter import *

from src.scheduling_and_allocation.gui_pages.views_controller_tk import ViewsControllerTk
from src.scheduling_and_allocation.model import Schedule, ResourceType, ConflictType

schedule = Schedule()

# teachers
t1 = schedule.add_update_teacher("Jane", "Doe", "0.25", teacher_id="1")
t2 = schedule.add_update_teacher("John", "Doe", teacher_id="2")
t3 = schedule.add_update_teacher("Babe", "Ruth", teacher_id="3")
t4 = schedule.add_update_teacher("Bugs", "Bunny", teacher_id="4")

# labs
l1 = schedule.add_update_lab("P107", "C-Lab")
l2 = schedule.add_update_lab("P322", "Mac Lab")
l3 = schedule.add_update_lab("P325")
l4 = schedule.add_update_lab("BH311", "Britain Hall")

# streams
st1 = schedule.add_update_stream("1A", "Math Stream")
st2 = schedule.add_update_stream("1B")

mw = Tk()
frame = Frame()
frame.pack(expand=1, fill="both")
resources:dict[ResourceType, list] = {ResourceType.teacher:[t1,t2,t3,t4,t1,t2,t3,t4,t1,t2,t3,t4,],
                           ResourceType.stream:[st1,st2],
                           ResourceType.lab:[l1,l2,l3,l4]}
vm = ViewsControllerTk(frame, resources, lambda *_:print(*_))
vm.set_button_colour(t1.number, ResourceType.teacher, ConflictType.TIME)
vm.set_button_colour(t2.number, ResourceType.teacher, ConflictType.NONE)
mw.mainloop()


# creating popup menu in tkinter
import tkinter
import os
import sys

from src.scheduling_and_allocation.gui_generics.menu_and_toolbars import MenuItem, MenuType, generate_menu

bin_dir: str = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(bin_dir, "../../"))


class A:
    # creates parent window
    def __init__(self):

        self.popup_menu = None
        self.root = tkinter.Tk()
        self.root.geometry('500x500')

        self.frame1 = tkinter.Label(self.root,
                                    width=400,
                                    height=400,
                                    text = "Hi",)
        self.frame1.pack()

    # create menu
    def popup(self):
        menu = list()
        menu.append(MenuItem(menu_type=MenuType.Command,
                                     label='say hi',
                                     command=lambda: self.hey("hi")
                                     )
                    )
        menu.append(MenuItem(menu_type=MenuType.Command,
                                     label='say hello',
                                     command=lambda: self.hey("hello")
                                     )
                    )
        menu.append(MenuItem(menu_type=MenuType.Command,
                                     label='say bye',
                                     command=lambda: self.hey("bye")
                                     )
                    )

        self.popup_menu = tkinter.Menu(self.root,
                                        tearoff=0)
        generate_menu(self.root, menu, self.popup_menu)

    # display menu on right click
    def do_popup(self, event):
        try:
            self.popup_menu.tk_popup(event.x_root,
                                     event.y_root)
        finally:
            self.popup_menu.grab_release()

    def hey(self, s):
        self.frame1.configure(text=s)

    def run(self):
        self.popup()
        self.root.bind("<Button-2>", self.do_popup)
        tkinter.mainloop()


a = A()
a.run()



from tkinter import *

from src.scheduling_and_allocation.gui_pages import DEColumnDescription, EditResourcesTk
from src.scheduling_and_allocation.modified_tk import set_default_fonts_and_colours


# TODO: button Colour changes don't work on mac, not yet tested on windows

def main():
    mw = Tk()
    colours, fonts = set_default_fonts_and_colours(mw, invert=True)
    mw.geometry('400x450')
    mw.title('Edit Resources Generic')
    frame = Frame(mw)
    frame.pack(expand=1, fill='both')

    col_descr = DEColumnDescription("Column 1", 30, "", False)
    col2_descr = DEColumnDescription("Column 2", 20, "", False)
    col3_descr = DEColumnDescription("Column 3", 15, "", False)

    columns = [col_descr, col2_descr, col3_descr]
    data = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    data2 = [[21, 22, 23], [24, 25, 26], [27, 28, 29]]

    de = EditResourcesTk(frame,
                         lambda *args: print("delete", [*args]),
                         lambda *args: print("save", [*args]),
                         colours,
                         )
    de.initialize_columns(columns)
    de.refresh(data)

    Button(mw, text="clear", command=lambda *_: de.data_entry.clear_data(),
           bg=colours.ButtonBackground,
           fg=colours.ButtonForeground,
           highlightbackground=colours.HighlightBackground,
           ).pack()
    Button(mw, text="refresh", command=lambda *_: de.refresh(data2)).pack()
    Button(mw, text="get data", command=lambda *_: print(de.get_all_data())).pack()

    print ("testing delete callback", de.delete_handler())
    print ("testing save callback", de.save_handler())
    mw.mainloop()


main()



from __future__ import annotations
import pytest
import sys

from src.scheduling_and_allocation.model import TimeSlot, Course, Section, Block, Lab, Teacher, ConflictType

course = Course("123")
dummy_Section = Section(course)


@pytest.fixture(scope="module", autouse=True)
def before_and_after_module():
    pass


@pytest.fixture(autouse=True)
def run_before():
    pass


def test_generic():
    day = 1
    start1 = 8
    start2 = 12
    dur = 2
    block1 = Block(dummy_Section, day, start1, dur, True)
    assert block1.section is dummy_Section
    assert not block1.conflict.is_conflicted()


def test_hashable():
    day = 1
    start1 = 8
    start2 = 12
    dur = 2
    block1 = Block(dummy_Section, day, start1, dur, True)
    block2 = Block(dummy_Section, day, start1, dur, True)
    s = {block1, block2}
    assert len(s) == 2


def test_sortable():
    slot1 = Block(dummy_Section, 2, start=13.25, duration=1.5)
    slot2 = Block(dummy_Section, 2, start=11.25, duration=1.5)
    s = [slot1, slot2]
    s.sort()
    assert s[0] == slot2
    assert s[1] == slot1


def test_sync_day():
    """Verifies that the day setter works as intended, changing the day property of any Block
    synced_blocks to the current Block. """
    day = 1
    start1 = 8
    start2 = 12
    dur = 2
    block1 = Block(dummy_Section, day, start1, dur, True)
    block2 = Block(dummy_Section, 4, start2, dur, True)
    block1.sync_block(block2)
    block2.sync_block(block1)
    assert block2._time_slot == block2._time_slot


def test_id():
    """Verifies that the id property works as intended."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur, True)
    old_id = block.id
    block = Block(dummy_Section, day, start, dur, True)
    assert block.id == old_id + 1


def test_id_with_id_given():
    """Verifies that the id property works as intended."""

    day = 1
    start = 8.5
    dur = 2
    existing_id = 12
    block1 = Block(dummy_Section, day, start, dur, True, block_id=existing_id)
    assert block1.id == existing_id
    block2 = Block(dummy_Section, day, start, dur, True)
    assert block2.id == existing_id + 1
    block3 = Block(dummy_Section, day, start, dur, True, block_id=existing_id - 5)
    assert block3.id == existing_id - 5
    block4 = Block(dummy_Section, day, start, dur, True)
    assert block4.id == block2.id + 1


def test_assign_lab_good():
    """Verifies that the add_lab method can assign a valid Lab object to the Block."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur, True)
    lab = Lab()
    block.add_lab(lab)
    assert block.has_lab(lab)


def test_assign_lab_multiple():
    """Verifies that the assign_lab_by_id method can assign a valid Lab object to the Block."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    lab1 = Lab("P100")
    lab2 = Lab("P101")
    block.add_lab(lab1)
    block.add_lab(lab2)
    labs = block.labs()
    assert block.has_lab(lab1)
    assert block.has_lab(lab2)
    assert len(labs) == 2


def test_assign_lab_multiple_with_duplicate():
    """Verifies that the assign_lab_by_id method can assign a valid Lab object to the Block."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    lab1 = Lab("P100")
    lab2 = Lab("p101")
    block.add_lab(lab1)
    block.add_lab(lab1)
    block.add_lab(lab2)
    labs = block.labs()
    assert len(labs) == 2


def test_remove_lab_good():
    """Verifies that remove_lab() can successfully remove a valid Lab object from this Block."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    lab1 = Lab("P100")
    lab2 = Lab("101")
    block.add_lab(lab1)
    block.add_lab(lab2)
    block.remove_lab(lab1)
    labs = block.labs()
    assert not block.has_lab(lab1)
    assert block.has_lab(lab2)
    assert len(labs) == 1


def test_remove_lab_no_crash():
    """Verifies that remove_lab won't crash the program if the Block doesn't contain
    the specified valid Lab. """
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    lab = Lab()
    block.add_lab(lab)
    other_lab = Lab("R-101", "dummy")
    block.remove_lab(other_lab)
    assert block.has_lab(lab)


def test_remove_all_labs():
    """Verifies that remove_all_labs works as intended, removing all Labs from the current
    Block. """
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    this_lab = Lab("R-100", "the second-worst place in the world")
    other_lab = Lab("R-101", "the worst place in the world")
    block.add_lab(this_lab)
    block.add_lab(other_lab)
    assert len(block.labs()) == 2
    block.remove_all_labs()
    assert len(block.labs()) == 0
    assert not block.has_lab(this_lab)
    assert not block.has_lab(other_lab)


def test_labs():
    """Verifies that labs returns a list of all Lab objects assigned to this Block."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    this_lab = Lab("R-100", "the second-worst place in the world")
    other_lab = Lab("R-101", "the worst place in the world")
    block.add_lab(this_lab)
    block.add_lab(other_lab)
    labs = block.labs()
    assert len(labs) == 2 and this_lab in labs and other_lab in labs


def test_labs_empty():
    """Verifies that lab_ids returns an empty list if called while no Labs are assigned to the
    Block. """
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    labs = block.labs()
    assert len(labs) == 0


def test_has_lab_good():
    """Verifies that has_lab_with_id() returns true when the specified Lab is assigned to this Block."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    lab = Lab()
    block.add_lab(lab)
    assert block.has_lab(lab)


def test_has_lab_false():
    """"Verifies that has_lab_with_id() returns false when the Lab isn't assigned to this Block."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    lab = Lab()
    assert block.has_lab(lab) is False


def test_assign_teacher_good():
    """Verifies that the add_teacher() method works as intended."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    teach = Teacher("John", "Smith")
    block.add_teacher(teach)
    assert block.has_teacher(teach)


def test_assign_teacher_multiple():
    """Verifies that the assign_teacher_by_id() method works as intended."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    teach1 = Teacher("John", "Smith")
    teach2 = Teacher("Jane", "Smith")
    block.add_teacher(teach1)
    block.add_teacher(teach2)
    assert block.has_teacher(teach1)
    assert block.has_teacher(teach2)
    assert len(block.teachers()) == 2


def test_assign_teacher_multiple_with_duplicates():
    """Verifies that the assign_teacher_by_id() method works as intended."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    teach1 = Teacher("John", "Smith")
    teach2 = Teacher("Jane", "Smith")
    block.add_teacher(teach1)
    block.add_teacher(teach1)
    block.add_teacher(teach2)
    print(teach1, teach2, block.teachers())
    assert len(block.teachers()) == 2


def test_remove_teacher_good():
    """Verifies that remove_teacher_by_id() works as intended."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    teach = Teacher("John", "Smith")
    block.add_teacher(teach)
    other_teach = Teacher("Jane", "Doe")
    block.add_teacher(other_teach)
    block.remove_teacher(teach)
    assert len(block.teachers()) == 1
    assert block.has_teacher(teach) is not True
    assert block.has_teacher(other_teach)


def test_remove_teacher_no_crash():
    """Verifies that remove_teacher doesn't crash the program when attempting to remove a
    Teacher that isn't assigned to this Block. """
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    teach = Teacher("John", "Smith")
    block.add_teacher(teach)
    other_teach = Teacher("Jane", "Doe")
    block.remove_teacher(other_teach)
    assert len(block.teachers()) == 1


def test_remove_all_teachers():
    """Verifies that remove_all_teachers() works as intended."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    teach = Teacher("John", "Smith")
    other_teach = Teacher("Jane", "Doe")
    block.add_teacher(teach)
    block.add_teacher(other_teach)
    block.remove_all_teachers()
    assert block.has_teacher(teach) is not True
    assert block.has_teacher(other_teach) is not True
    assert len(block.teachers()) == 0


def test_teachers():
    """Verifies that teacher_ids() returns a list of all the Teachers assigned to this Block."""
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    teach = Teacher("John", "Smith")
    other_teach = Teacher("Jane", "Doe")
    block.add_teacher(teach)
    block.add_teacher(other_teach)
    teachers = block.teachers()
    assert len(teachers) == 2 and teach in teachers and other_teach in teachers


def test_teachers_empty_list():
    """Verifies that teacher_ids will return an empty list if no Teachers are assigned to this
    Block. """
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    teachers = block.teachers()
    assert len(teachers) == 0


def test_has_teacher_true():
    """Verifies that has_teacher returns true when the specified Teacher has been assigned
    to this Block. """
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    teach = Teacher("John", "Smith")
    block.add_teacher(teach)
    assert block.has_teacher(teach)


def test_has_teach_false():
    """Verifies that has_teacher returns false when the specified Teacher isn't assigned to
    this Block. """
    day = 1
    start = 8.5
    dur = 2
    block = Block(dummy_Section, day, start, dur)
    teach = Teacher("John", "Smith")
    block.add_teacher(teach)
    other_teach = Teacher("Jane", "Doe")
    assert block.has_teacher(other_teach) is False


def test_reset_conflicted():
    """Verifies that reset_conflicted works as intended, setting the value of conflicted_number to
    False. """
    day = 1
    start = 8.5
    dur = 2

    block1 = Block(dummy_Section, day, start, dur)
    block1.clear_conflicts()
    assert not block1.conflict.is_conflicted()


def test_conflicted_getter():
    """Verifies that the conflicted_number getter works as intended, returning a default value of
    False. """
    day = 1
    start = 8.5
    dur = 2
    block1 = Block(dummy_Section, day, start, dur)
    assert not block1.conflict.is_conflicted()


def test_conflicted_setter_good():
    """Verifies that the conflicted_number setter works as intended."""
    day = 1
    start = 8.5
    dur = 2
    block1 = Block(dummy_Section, day, start, dur)
    block1.clear_conflicts()
    block1.add_conflict(ConflictType.LUNCH)
    assert block1.conflict == ConflictType.LUNCH


def test_is_conflicted_true():
    """Verifies that is_conflicted() returns true if the conflicted number is not zero."""
    day = 1
    start = 8.5
    dur = 2
    block1 = Block(dummy_Section, day, start, dur)
    block1.add_conflict(ConflictType.LUNCH)
    assert block1.conflict.is_conflicted()


def test_is_conflicted_false():
    """Verifies that is_conflicted() returns the false if the conflicted number is zero."""
    day = 1
    start = 8.5
    dur = 2

    block1 = Block(dummy_Section, day, start, dur)
    block1.add_conflict(ConflictType.LUNCH)
    assert block1.conflict.is_conflicted()
    block1.clear_conflicts()
    assert not block1.conflict.is_conflicted()


def test_adjusting_conflict_sets_appropriate_bits():
    """Verifies that is_conflicted() returns the false if the conflicted number is zero."""
    day = 1
    start = 8.5
    dur = 2
    block1 = Block(dummy_Section, day, start, dur)
    block1.clear_conflicts()
    block1.add_conflict(ConflictType.LUNCH)
    block1.add_conflict(ConflictType.LUNCH)
    block1.add_conflict(ConflictType.TIME_LAB)
    block1.add_conflict(ConflictType.TIME_LAB)
    assert block1.conflict.is_time_lab()
    assert block1.conflict.is_time_lunch()


# ====================================================================================================================
# Synchronization
# ====================================================================================================================
def test_start_setter_synced_2_blocks_check_block1_moves_block2():
    day = 1
    start = 8.5
    dur = 2
    block1 = Block(dummy_Section, day, start, dur)
    block2 = Block(dummy_Section, 3, start, dur)
    block1.sync_block(block2)
    assert block2._time_slot == block2._time_slot


def test_start_setter_synced_2_blocks_check_block2_moves_block1():
    day = 1
    start = 8.5
    dur = 2
    block1 = Block(dummy_Section, day, start, dur)
    block2 = Block(dummy_Section, 3, start, dur)
    block1.sync_block(block2)
    assert block2._time_slot == block2._time_slot
    block1._time_slot.day = 2
    assert block2._time_slot == block2._time_slot


def test_start_setter_synced_4_blocks():
    """Same as previous test, but with 4 blocks instead of 2."""
    day = 1
    start = 8.5
    dur = 2
    block1 = Block(dummy_Section, day, start, dur)
    block2 = Block(dummy_Section, 3, start, dur)
    block3 = Block(dummy_Section, 4, start, dur)
    block4 = Block(dummy_Section, 5, start, dur)
    block1.sync_block(block2)
    block1.sync_block(block3)
    block1.sync_block(block4)

    new_start = "10:00"
    block1._time_slot.time_start = new_start
    assert block2._time_slot == block3._time_slot == block4._time_slot


def test_start_setter_synced_4_blocks_move_block3():
    """Same as previous test, but with 4 blocks instead of 2."""
    day = 1
    start = 8.5
    dur = 2
    block1 = Block(dummy_Section, day, start, dur)
    block2 = Block(dummy_Section, 3, start, dur)
    block3 = Block(dummy_Section, 4, start, dur)
    block4 = Block(dummy_Section, 5, start, dur)
    block1.sync_block(block2)
    block1.sync_block(block3)
    block1.sync_block(block4)

    new_start = "10:00"
    block3._time_slot.time_start = new_start
    assert block2._time_slot.time_start == new_start == block3._time_slot.time_start == block4._time_slot.time_start


def test_sync_block_good():
    """Verifies that sync_block() synchronizes a passed Block with this Block."""
    day = 1
    start = 8.5
    dur = 2
    block1 = Block(dummy_Section, day, start, dur)
    block2 = Block(dummy_Section, 2, start, dur)
    block1.sync_block(block2)
    assert block2 in block1.synced_blocks()
    assert block1 in block2.synced_blocks()


def test_unsync_block1_from_block2():
    """Verifies that unsync_block() works as intended."""
    day = 1
    start = 8.5
    dur = 2
    block1 = Block(dummy_Section, day, start, dur)
    block2 = Block(dummy_Section, 2, start, dur)
    block1.sync_block(block2)
    assert block1 in block2.synced_blocks()
    assert block2 in block1.synced_blocks()

    block2.unsync_block(block1)
    assert block1 not in block2.synced_blocks()
    assert block2 not in block1.synced_blocks()


def test_synced():
    """Verifies that synced_blocks returns a tuple of all the Blocks that are synced_blocks with this
    Block. """
    day = 1
    start = 8.5
    dur = 2
    block1 = Block(dummy_Section, day, start, dur)
    block2 = Block(dummy_Section, 2, start, dur)
    block3 = Block(dummy_Section, 3, start, dur)
    block1.sync_block(block2)
    block1.sync_block(block3)
    blocks = block1.synced_blocks()

    assert len(blocks) == 2 and block2 in blocks and block3 in blocks


def test_synced_empty_tuple():
    """Verifies that synced_blocks() returns an empty tuple if called when no Blocks are synced_blocks with
    this Block. """
    day = 1
    start = 8.5
    dur = 2
    block1 = Block(dummy_Section, day, start, dur)
    blocks = block1.synced_blocks()
    assert len(blocks) == 0


def test_synced_blocks_interdependence_works_as_expected():
    block1 = Block(dummy_Section, 1, 8.5, 2)
    block2 = Block(dummy_Section, 2, 8.5, 2)
    block3 = Block(dummy_Section, 3, 8.5, 2)
    block4 = Block(dummy_Section, 5, 8.5, 2)

    # make all blocks dependent on block1, and vice versa
    block1.sync_block(block2)
    block1.sync_block(block3)
    block1.sync_block(block4)

    # disconnect blocks 1 from blocks 3,
    block3.unsync_block(block1)

    # changing blocks 3 should affect on anybody block2 and block4, but not block1
    block3._time_slot.time_start = 11.5
    assert block1._time_slot == TimeSlot(1, 8.5, 2)
    assert block3._time_slot.time_start == block2._time_slot.time_start == block4._time_slot.time_start == 11.5

    # changing block 1 should not affect any other block
    block1._time_slot.time_start = 9.5
    assert block3._time_slot.time_start == block2._time_slot.time_start == block4._time_slot.time_start == 11.5
    assert block1._time_slot.time_start == 9.5
